"""
The main Zwiki module. See README.txt.

(c) 1999-2002 Simon Michael <simon@joyful.com> for the zwiki community.
Wikiwikiweb formatting by Tres Seaver <tseaver@zope.com>
Parenting code and regulations by Ken Manheimer <klm@zope.com>
Initial Zope CMF integration by Chris McDonough <chrism@zope.com>
Full credits are at http://zwiki.org/ZwikiContributors .

This product is available under the GNU GPL.  All rights reserved, all
disclaimers apply, etc.
"""

__version__="$Revision: 0.277 $"[11:-2]

import os, sys, re, string, time, math, traceback
from DateTime import DateTime
from string import split,join,find,lower,rfind,atoi,strip,replace
from types import *
from wwml import translate_WMML
from AccessControl import getSecurityManager
import Acquisition
from App.Common import absattr, rfc1123_date, aq_base
import DocumentTemplate
from urllib import quote, unquote
import ZODB
import Globals
from Globals import HTMLFile, MessageDialog, package_home
from Products.MailHost.MailHost import MailBase
from OFS.content_types import guess_content_type
from OFS.Document import Document
from OFS.DTMLDocument import DTMLDocument
import OFS.Image
import StructuredText

from Products.ZWiki import __version__
from Defaults import DEFAULT_PAGE_TYPE, DISABLE_JAVASCRIPT, \
     LARGE_FILE_SIZE, AUTO_UPGRADE, \
     default_standard_wiki_header, default_standard_wiki_footer, \
     default_editform, default_backlinks, default_subscribeform, \
     default_wikipage
from Diff import DiffSupport
from Parents import ParentsSupport
import Permissions
from Regexps import url, bracketedexpr, footnoteexpr, wikiname1, wikiname2, \
     simplewikilink, wikilink, interwikilink, remotewikiurl, protected_line, \
     javascriptexpr, dtmlorsgmlexpr, zwikiidcharsexpr, anywikilinkexpr, \
     localwikilink, spaceandlowerexpr, untitledwikilinkexpr, htmlheaderexpr, \
     htmlfooterexpr
from Utils import thunk_substituter, within_literal, html_quote, html_unquote,\
     DLOG, myDocument, ZOPEVERSION, _, N_, withinSgmlOrDtml, parseHeadersBody
from Mail import MailSupport
from CatalogAwareness import CatalogAwareness
from Tracker import TrackerSupport
from Regulations import RegulationsSupport
try:
    #XXX fails when running unit tests, for some reason
    from CMF import CMFAwareness
except ImportError:
    class CMFAwareness:
        pass

###########################################################################
# CLASS ZwikiPage
###########################################################################

class ZWikiPage(
    DTMLDocument,
    ParentsSupport,
    DiffSupport,
    MailSupport,
    CatalogAwareness,
    TrackerSupport,
    RegulationsSupport,
    CMFAwareness,
    ):
    """
    A ZWikiPage is a DTML Document which knows how to render itself in
    various wiki styles, and provides some utility methods to support
    wiki-building.

    RESPONSIBILITIES:

      - render itself

      - provide edit/create forms, backlinks, table of contents

      - accept edit/create requests, with authentication

      - store metadata such as permissions, time, last author, parents etc

      - manage subscriber lists for self & parent folder

    """

    ######################################################################
    # attributes

    #XXX calculate dynamically ?
    ZWIKI_PAGE_TYPES = (
        'stxprelinkdtmlhtml',
        'stxdtmllinkhtml',
        'dtmlstxlinkhtml',
        'stxprelinkhtml',
        'stxlinkhtml',
        'stxprelink',
        'stxlink',
        'wwmlprelink',
        'wwmllink',
        'prelinkdtmlhtml',
        'dtmllinkhtml',
        'prelinkhtml',
        'linkhtml',
        'dtmlhtml',
        'html',
        'textlink',
        'plaintext',
    )

    meta_type = "ZWiki Page"
    icon      = "misc_/ZWiki/ZWikiPage_icon"
    page_type = DEFAULT_PAGE_TYPE
    creator = ''
    creator_ip = ''
    creation_time = ''
    last_editor = ''
    last_editor_ip = ''
    last_edit_time = ''
    last_log = ''
    _prerendered = '' # cached partially rendered text
    _prelinked = []     # cached links & rendered text regions
    _links = []         # cached unique links

    # properties visible in the ZMI
    # would rather append to the superclass' _properties here
    # DocumentTemplate.inheritedAttribute('_properties'),...) ?
    _properties=(
        {'id':'title', 'type': 'string', 'mode':'w'},
        {'id':'page_type', 'type': 'selection', 'mode': 'w',
         'select_variable': 'ZWIKI_PAGE_TYPES'}, # shows all types
        {'id':'creator', 'type': 'string', 'mode': 'r'},
        {'id':'creator_ip', 'type': 'string', 'mode': 'r'},
        {'id':'creation_time', 'type': 'string', 'mode': 'r'},
        {'id':'last_editor', 'type': 'string', 'mode': 'r'},
        {'id':'last_editor_ip', 'type': 'string', 'mode': 'r'},
        {'id':'last_edit_time', 'type': 'string', 'mode': 'r'},
        {'id':'last_log', 'type': 'string', 'mode': 'r'},
        ) \
        + ParentsSupport._properties \
        + MailSupport._properties \
        + CatalogAwareness._properties

    # permissions
    # Most of these we check in the code ourselves, because edit()
    # does a lot of different things. This also allows more helpful
    # authorization errors.
    __ac_permissions__=(
        (Permissions.Append, ()),
        (Permissions.ChangeType, ()),
        (Permissions.Change, ('PUT','manage_edit','manage_upload','editform')),
        (Permissions.Rename, ('rename')),
        (Permissions.Delete, ()),
        (Permissions.FTP, ('manage_FTPstat','manage_FTPget','manage_FTPlist')),
        ) \
        + ParentsSupport.__ac_permissions__ \
        + RegulationsSupport.__ac_permissions__

    ######################################################################
    # initialization methods

    def __init__(self, source_string='', mapping=None, __name__=''):
        """
        Initialise this instance, including it's CMF data if applicable.

        Ugly, but putting CMFAwareness before DTMLDocument in the
        inheritance order creates problems.
        """
        if hasattr(CMFAwareness,'__init__'):
            CMFAwareness.__init__(self,
                                  source_string=source_string,
                                  mapping=mapping,
                                  __name__=__name__,
                                  )
        else:
            DTMLDocument.__init__(self,
                                  source_string=source_string,
                                  mapping=mapping,
                                  __name__=__name__,
                                  )

    ######################################################################
    # rendering methods

    def __call__(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """Render this zwiki page, upgrading it on the fly if needed
        """
        if AUTO_UPGRADE: self.upgrade()
        body = apply(self._render,(client,REQUEST,RESPONSE),kw)
	if RESPONSE is not None:
	    RESPONSE.setHeader('Content-Type', 'text/html')
            #RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime)) 
            #causes browser caching problems ? 
        return body

    def _render(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """
        Render the body of this zwiki page according to it's page_type
        """
        method = self._renderMethod()
        return apply(method,(self, REQUEST, RESPONSE), kw)

    def _renderMethod(self):
        r = 'render_' + self.page_type
        if hasattr(self, r):
            method = getattr(self,r)
        else:
            method = self.render_plaintext
        return method

    def _preRender(self,clear_cache=0):
        """
        Make sure any applicable pre-rendering for this page has been done
        
        each render_* method knows how to do it's own prerendering..
        pass the pre_only flag in kw to preserve the standard dtml method
        signature
        if clear_cache is 1, blow away any cached data
        """
        if clear_cache:
            self.clearCache()
        return apply(self._render,(self, {}, None), {'pre_only':1})

    def clearCache(self,REQUEST=None):
        """
        forcibly clear out any cached render data for this page
        """
        self._prerendered = ''
        self._prelinked = []
        self._links = []
        if REQUEST:
            REQUEST.RESPONSE.redirect(self.page_url())

    def addStandardLayoutTo(self,body,**kw):
        """
        Add standard wiki page layout to the rendered page body.

        Tries these alternatives:
        
        1. if a wikipage page template is found in the zodb (in this
           folder or acquired, use that

        2. if a standard_wiki_header or standard_wiki_footer dtml method
        is found in the zodb, use that. If one is missing, use the default
        value for the other. Note the defaults for these methods were last
        updated just before 0.10.0.
        
        3. otherwise, use the default page template from the filesystem.
.
        Also checks for bare/noheader/nofooter keywords.
        """
        REQUEST = getattr(self,'REQUEST',None)
        RESPONSE = getattr(REQUEST,'RESPONSE',None)
        folder = self.folder()
        if (hasattr(folder,'wikipage') and
            not hasattr(REQUEST,'bare') and
            not kw.has_key('bare') and
            getattr(folder.wikipage,'meta_type',None) == 'Page Template'):
            if kw:
                kw['body'] = body
            else:
                kw = {'body':body}
            return apply(folder.wikipage.__of__(self),(),kw)
        elif ((hasattr(folder,'standard_wiki_header') and
               getattr(folder.standard_wiki_header,
                       'meta_type',None) == 'DTML Method') or
              (hasattr(folder,'standard_wiki_footer') and
               getattr(folder.standard_wiki_footer,
                       'meta_type',None) == 'DTML Method')):
            header = apply(self._renderHeaderOrFooter,('header',REQUEST,RESPONSE),kw)
            footer = apply(self._renderHeaderOrFooter,('footer',REQUEST,RESPONSE),kw)
            return header + body + footer
        elif (self.wikipage and
              not hasattr(REQUEST,'bare') and
              not kw.has_key('bare')):
            if kw:
                kw['body'] = body
            else:
                kw = {'body':body}
            #return apply(default_wikipage.__of__(self),(),kw)
            return apply(self.wikipage,(),kw)
        else:
            return body

    def _renderHeaderOrFooter(self, hdrOrFtr, REQUEST, RESPONSE, **kw):
        """
        Generate the standard wiki header or footer for this page.

        Also check for various options which may disable this.
        hdrOrFtr should be 'header' or 'footer'
        """
        # actually a flag in REQUEST is not much use, try a keyword arg
        # why doesn't <dtml-var WikiPage bare> work yet ?
        if (REQUEST is not None and
            not (hasattr(REQUEST,'bare') or
                 hasattr(REQUEST,'no'+hdrOrFtr) or
                 kw.has_key('bare') or
                 kw.has_key('no'+hdrOrFtr)) and
            (hdrOrFtr == 'header' or hdrOrFtr == 'footer')):
            method = getattr(self,'standard_wiki_'+hdrOrFtr)
            #return method(self, REQUEST, RESPONSE, **kw)
            return apply(method,(self, REQUEST, RESPONSE), kw)
        else:
            return ''

    # override cook so DTML will use our pre-rendered data if available
    import thread
    def cook(self,
             cooklock=thread.allocate_lock(),
             ):
        cooklock.acquire()
        try:
            self._v_blocks=self.parse(self._prerendered or self.read())
            self._v_cooked=None
        finally:
            cooklock.release()

    # built-in render methods (page types)
    #
    # add new ones as dtml/python/external methods. I inlined all these
    # for maximum control. Perhaps these should become objects, but no
    # hurry; they still have the signature of a DTML method, which seems
    # possibly useful
    #
    # for 0.9.10, added pre-rendering (and pre-linking)
    # same goal as in WikiForNow, different implementation
    # each render method does as much pre-rendering as it can and caches
    # it in _prerendered; this is updated if needed, and also when the
    # page text or type is changed.
    # this allows flexibility for the different page types and
    # keeps the rendering code together
    # NB _prerendered == '' is assumed to mean this page has not been
    # pre-rendered, which may not be true
    # All render methods may be called with pre_only:1 in kw to do only
    # the pre-rendering
    #
    # Could this be simpler ? Perhaps. We are exploring a range of
    # combinations of rendering rules and strategies, while meeting the
    # constraints of DTML, Structured Text etc. and aiming for blazing
    # performance. It's a bit of a dance.

    def render_stxprelinkdtmlhtml(self, client=None, REQUEST={},
                                  RESPONSE=None, **kw):
        """
        render the page using structured text + wiki links + DTML + HTML

        As well as stx formatting, this one does pre-linking - as much as
        possible of the linking work - ahead of time.

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='stxprelinkdtmlhtml')
        >>> p._render(bare=1)
        '<p> PageOne PageTwo</p>\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = self.applyLineEscapesIn(t)
            t = self.stxToHtml(t)
            self._prerendered = t or '\n'
            self.cook()      # pre-parse DTML
            self._preLink()  # pre-parse links
        if kw.get('pre_only',0): return
        # final render
        self._renderWithLinks()
        t = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)
        t = self._renderLateLinkTitles(t)
        #t = self.addStandardLayoutTo(t,**kw)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_stxdtmllinkhtml(self, client=None, REQUEST={},
                               RESPONSE=None, **kw):
        """
        render the page using structured text + DTML + wiki links + HTML

        Behaviour change since 0.9.9: switches stx/dtml rendering order.
        This one caches the structured text formatting ahead of time.

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='stxdtmllinkhtml')
        >>> p._render(bare=1)
        '<p> PageOne PageTwo</p>\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = self.applyLineEscapesIn(t)
            t = self.stxToHtml(t)
            self._prerendered = t or '\n'
            self.cook()
        if kw.get('pre_only',0): return
        # final render
        t = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)
        t = self.renderLinksIn(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_dtmlstxlinkhtml(self, client=None, REQUEST={},
                               RESPONSE=None, **kw):
        """
        render the page using DTML + structured text + wiki links + HTML

        This is the old DTML-first non-caching version used up till 0.9.9.

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='dtmlstxlinkhtml')
        >>> p._render(bare=1)
        '<p> PageOne PageTwo</p>\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            self._prerendered = t or '\n'
            self.cook()
        if kw.get('pre_only',0): return
        # final render
        t = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)
        t = self.applyLineEscapesIn(t)
        t = self.stxToHtml(t)
        t = self.renderLinksIn(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_stxprelinkhtml(self, client=None, REQUEST={},
                              RESPONSE=None, **kw):
        """
        render the page using structured text + wiki links + HTML, prelinking

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='stxprelinkhtml')
        >>> p._render(bare=1)
        '<p> PageOne PageTwo</p>\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = self.applyLineEscapesIn(t)
            t = self.stxToHtml(t)
            self._prerendered = t or '\n'
            self._preLink() 
        if kw.get('pre_only',0): return
        # final render
        self._renderWithLinks()
        t = self._prerendered
        t = self._renderLateLinkTitles(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_stxlinkhtml(self, client=None, REQUEST={},
                           RESPONSE=None, **kw):
        """
        render the page using structured text + wiki links + HTML

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='stxlinkhtml')
        >>> p._render(bare=1)
        '<p> PageOne PageTwo</p>\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = self.applyLineEscapesIn(t)
            t = self.stxToHtml(t) or '\n'
            self._prerendered = t
        if kw.get('pre_only',0): return
        # final render
        t = self._prerendered
        t = self.renderLinksIn(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_stxprelink(self, client=None, REQUEST={},
                          RESPONSE=None, **kw):
        """
        render the page using structured text + wiki links, prelinking

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='stxprelink')
        >>> p._render(bare=1)
        '<p> PageOne PageTwo</p>\\n'
        >>> 
        >>> p.edit(text='"link containing WikiName":url\\n',type='stxprelink')
        >>> p._render(bare=1)
        '<p><a href="url">link containing WikiName</a></p>\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = html_quote(t)
            t = self.applyLineEscapesIn(t)
            t = self.stxToHtml(t)
            self._prerendered = t or '\n'
            self._preLink() 
        if kw.get('pre_only',0): return
        # final render
        self._renderWithLinks()
        t = self._prerendered
        t = self._renderLateLinkTitles(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_stxlink(self, client=None, REQUEST={},
                       RESPONSE=None, **kw):
        """
        render the page using structured text + wiki links

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='stxlink')
        >>> p._render(bare=1)
        '<p> PageOne PageTwo</p>\\n'
        >>> 
        >>> p.edit(text='"link text":url\\n',type='stxlink')
        >>> p._render(bare=1)
        '<p><a href="url">link text</a></p>\\n'
        >>>
        
        #>>> p.edit(text='"link containing WikiName":url\\n',type='stxlink')
        #>>> p._render(bare=1)
        #'<p><a href="url">link containing WikiName</a></p>\\n'
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = html_quote(t)
            t = self.applyLineEscapesIn(t)
            t = self.stxToHtml(t)
            self._prerendered = t or '\n'
        if kw.get('pre_only',0): return
        # final render
        t = self._prerendered
        t = self.renderLinksIn(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_prelinkdtmlhtml(self, client=None, REQUEST={},
                               RESPONSE=None, **kw):
        """
        render the page using DTML + wiki links + HTML, prelinking

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='prelinkdtmlhtml')
        >>> p._render(bare=1)
        ' PageOne PageTwo\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = self.applyLineEscapesIn(t)
            self._prerendered = t or '\n'
            self.cook()
            self._preLink()
        if kw.get('pre_only',0): return
        # final render
        self._renderWithLinks()
        t = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)
        t = self._renderLateLinkTitles(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_dtmllinkhtml(self, client=None, REQUEST={},
                            RESPONSE=None, **kw):
        """
        render the page using DTML + wiki links + HTML

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='dtmllinkhtml')
        >>> p._render(bare=1)
        ' PageOne PageTwo\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = self.applyLineEscapesIn(t)
            self._prerendered = t or '\n'
            self.cook()
        if kw.get('pre_only',0): return
        # final render
        t = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)
        t = self.renderLinksIn(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_prelinkhtml(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """
        render the page using wiki links + HTML, prelinking

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='prelinkhtml')
        >>> p._render(bare=1)
        ' PageOne PageTwo\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = self.applyLineEscapesIn(t)
            self._prerendered = t or '\n'
            self._preLink()
        if kw.get('pre_only',0): return
        # final render
        self._renderWithLinks()
        t = self._prerendered
        t = self._renderLateLinkTitles(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_linkhtml(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """
        render the page using wiki links + HTML

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='linkhtml')
        >>> p._render(bare=1)
        ' PageOne PageTwo\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = self.applyLineEscapesIn(t)
            self._prerendered = t or '\n'
        if kw.get('pre_only',0): return
        # final render
        t = self._prerendered
        t = self.renderLinksIn(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_dtmlhtml(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """
        render the page using DTML + HTML, nothing else

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='dtmlhtml')
        >>> p._render(bare=1)
        '! PageOne PageTwo\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            self._prerendered = t or '\n'
            self.cook()
        if kw.get('pre_only',0): return
        # final render
        t = apply(DTMLDocument.__call__,(self, client, REQUEST, RESPONSE), kw)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_html(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """
        render the page using HTML with no surprises

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='dtmlhtml')
        >>> p._render(bare=1)
        '! PageOne PageTwo\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            self._prerendered = t or '\n'
        if kw.get('pre_only',0): return
        # final render
        t = self._prerendered
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_wwmlprelink(self, client=None, REQUEST={},
                           RESPONSE=None, **kw):
        """
        render the page with WikiWikiWeb formatting + (z)wiki links, prelinking

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='wwmlprelink')
        >>> p._render(bare=1)
        '    PageOne PageTwo\\n<P>'

        # why pre ?
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = html_quote(t)
            t = self.applyLineEscapesIn(t)
            t = translate_WMML(t)
            self._prerendered = t or '\n'
            self._preLink()
        if kw.get('pre_only',0): return
        # final render
        self._renderWithLinks()
        t = self._prerendered
        t = self._renderLateLinkTitles(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_wwmllink(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """
        render the page using WikiWikiWeb formatting + (z)wiki links

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='wwmllink')
        >>> p._render(bare=1)
        '    PageOne PageTwo\\n<P>'
        
        #why do we get <pre> here ?
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = html_quote(t)
            t = self.applyLineEscapesIn(t)
            t = translate_WMML(t)
            self._prerendered = t or '\n'
        if kw.get('pre_only',0): return
        # final render
        t = self._prerendered
        t = self.renderLinksIn(t)
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_textlink(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """
        render the page using fixed-width plain text plus zwiki links
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = html_quote(t)
            self._prerendered = t or '\n'
        if kw.get('pre_only',0): return
        # final render
        t = self._prerendered
        t = self.renderLinksIn(t)
        t = "<pre>\n" + t + "\n</pre>\n"
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def render_plaintext(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """
        render the page using fixed-width plain text with no surprises

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> 
        >>> p.edit(text='! PageOne PageTwo\\n',type='plaintext')
        >>> p._render(bare=1)
        '<pre>\\n! PageOne PageTwo\\n\\n</pre>\\n'
        
        """
        # pre render
        t = str(self.read())
        if not self._prerendered:
            get_transaction().note('prerender')
            t = "<pre>\n" + html_quote(t) + "\n</pre>\n"
            self._prerendered = t or '\n'
        if kw.get('pre_only',0): return
        # final render
        t = self._prerendered
        t = apply(self.addStandardLayoutTo,(t,),kw)
        return t

    def stxToHtml(self, text):
        """
        Render some text into html according to the structured text rules,
        along with our workarounds and customizations
        """
        text = str(text)        
        if ZOPEVERSION < (2,4):
            # zope 2.2 - 2.3.x, classic structured text

            # final single-line paragraph becomes a heading if there are
            # trailing blank lines - strip them
            text = re.sub(r'(?m)\n[\n\s]*$', r'\n', text)

            # an initial single word plus period becomes a numeric bullet -
            # prepend a temporary marker to inhibit
            text = re.sub(r'(?m)^([ \t]*)([A-z]\w*\.)',
                          r'\1<!--NOSTX-->\2',
                          text)

            # :: quoting fails if there is whitespace after the :: - remove it
            text = re.sub(r'(?m)::[ \t]+$', r'::', text)

            # suppress stx footnote handling so we can do it our way
            text = re.sub(footnoteexpr,r'<a name="ref\1">![\1]</a>',text)
            text = re.sub(r'(?m)\[',r'[<!--NOSTX-->',text)

            # generate html
            text = str(StructuredText.HTML(text,level=3))

        else:
            # zope >= 2.4, structured text NG

            # an initial single word plus period becomes a numeric bullet -
            # prepend a temporary marker to inhibit
            text = re.sub(r'(?m)^([ \t]*)([A-z]\w*\.)',
                          r'\1<!--NOSTX-->\2',
                          text)

            # :: quoting fails if there is whitespace after the :: - remove it
            text = re.sub(r'(?m)::[ \t]+$', r'::', text)

            # suppress stx footnote handling so we can do it our way
            text = re.sub(footnoteexpr,r'<a name="ref\1">![\1]</a>',text)
            text = re.sub(r'(?m)\[',r'[<!--NOSTX-->',text)

            # generate html
            text = StructuredText.HTMLNG(
                myDocument(StructuredText.Basic(str(text))),
                level=3)

        # clean up
        text = re.sub(r'(<|&lt;)!--NOSTX--(>|&gt;)', r'', text)

        # strip html & body if added (not all zope versions support header=0)
        text = re.sub(
            r'(?sm)^<html.*<body.*?>\n(.*)</body>\n</html>\n',r'\1',text)

        return text

    def _renderLateLinkTitles(self,text):
        """
        fill in titles for any untitled wiki links in text
    
        In pre-linking modes, we do this in a final pass after linking &
        dtml so that we can have fast-changing link titles without messing
        up our caching. See _renderWithLinks for more.
        #    The regexp needs to match _renderLink's link format.  We can
        assume the link will be all on one line and that there's no raw
        dtml lying around.
        """
        return re.sub(untitledwikilinkexpr,self._replaceUntitledWikilink,text)

    def _replaceUntitledWikilink(self, match):
        """
        Replace a untitledwikilink occurence with a suitably titled link.
        """
        link = match.group()
        pagename = match.group('page')
        # old wiki page urls may need un-quoting to get the id
        # not quite sure how to handle this correctly
        #try:
        #    page = self.folder()[pagename]
        #except KeyError:
        #    page = self.folder()[unquote(pagename)]
        # I think they all need unquoting
        # also allow for differing wikiname & id now
        page = self.pageWithName(unquote(pagename))
        # bah!!!
        if page:
            linktitle = page.linkTitle()
        else:
            linktitle = "XXX zwiki bug - couldn't find page"
        newlink = re.sub(r' title="">', r' title="%s">' % (linktitle), link)
        return newlink

    def supportsStx(self):
        """does this page do Structured Text formatting ?"""
        return re.search(r'(?i)(structuredtext|stx)',
                         self.page_type) is not None

    def supportsWiki(self):
        """does this page do wiki linking ?"""
        return re.search(r'(?i)plain',self.page_type) is None

    def supportsHtml(self):
        """does this page render embedded HTML ?"""
        return not self.page_type in \
               ['structuredtextonly','classicwiki','plaintext']

    def supportsDtml(self):
        """does this page support embedded DTML ?"""
        return re.search(r'(?i)dtml',self.page_type) is not None

    def hasDynamicContent(self):
        """does this page contain dynamic content ?"""
        return (self.supportsDtml() and
                re.search(r'(?i)(<!--|<dtml|&dtml)',self.read()) is not None)

    def supportsPreLinking(self):
        """does this page do prelinking ?"""
        return re.search(r'prelink',self.page_type) is not None

    # from WFN - not used yet, may come in handy
    def prep_citation(self, rfind=string.rfind, strip=string.strip):
        """Quote text for use in literal citations.

        We prepend '>' to each line, splitting long lines (propagating
        existing citation and leading whitespace) when necessary.
        """
        got = []
        for line in string.split(self._st_data or self.xread(), '\n'):
            pref = '> '
            if len(line) < 79:
                got.append(pref + line)
                continue
            m = cite_prefixexp.match(line)
            if m is None:
                pref = '> %s'
            else:
                if m.group(1):
                    pref = pref + m.group(1)
                    line = line[m.end(1)+1:]
                    if m.end(1) > 60:
                        # Too deep quoting - collapse it:
                        pref = '> >> '
                        lencut = 0
                pref = pref + '%s'
                leading_space = m.group(2)
                if leading_space:
                    pref = pref + leading_space
                    line = line[len(leading_space):]
            lenpref = len(pref)
            continuation_padding = ''
            lastcurlen = 0
            while 1:
                curlen = len(line) + lenpref
                if curlen < 79 or (lastcurlen and lastcurlen <= curlen):
                    # Small enough - we're done - or not shrinking - bail out
                    if line: got.append((pref % continuation_padding) + line)
                    break
                else:
                    lastcurlen = curlen
                splitpoint = max(rfind(line[:78-lenpref], ' '),
                                 rfind(line[:78-lenpref], '\t'))
                if not splitpoint or splitpoint == -1:
                    if strip(line):
                        got.append((pref % continuation_padding) +
                                   line)
                    line = ''
                else:
                    if strip(line[:splitpoint]):
                        got.append((pref % continuation_padding) +
                                   line[:splitpoint])
                    line = line[splitpoint+1:]
                if not continuation_padding:
                    # Continuation lines are indented more than intial - just
                    # enough to line up past, eg, simple bullets.
                    continuation_padding = '  '
        return string.join(got, '\n')

    ######################################################################
    # linking & link rendering methods
    # XXX refactoring in progress. A couple of ways to get stuff linked:
    #
    # 1. the old way
    # _oldRenderLinksIn(text) or wikilink(text) - called at view time
    #  replaces interwiki links in text
    #   calls _replaceInterWikilink for each
    #  replaces other links in text
    #   calls _replaceWikilink for each
    #
    # 2.1 the new way (link arbitrary text)
    # renderLinksIn(text) - called at view time
    #  replaces links in text
    #   calls _renderLink for each
    #
    # 2.2 the new way (link this page's text with caching)
    # _preLink() - called at edit time
    #  searches for valid links in self._prerendered
    #  caches them with text regions in self._prelinked (also self._links)
    # _renderWithLinks() - called at view time
    #  calls _renderedLinks()
    #   calls _renderLink for each cached link
    #    calls _replaceInterWikilink & emulates _replaceWikilink
    #  interpolates rendered links with cached text regions
    #  saves final text back in self._prerendered (dtml: re-cook if changed)

    def applyLineEscapesIn(self, text):
        """
        implement wikilink-escaping in lines in text which begin with !
        """
        return re.sub(protected_line, self._protectLine, text)
        
    def _protectLine(self, match):
        """
        return the string represented by match with all it's wikilinks escaped
        """
        return re.sub(wikilink, r'!\1', match.group(1))

    def _oldRenderLinksIn(self,text,doLineEscapes=0):
        """
        Return text with all links rendered as for the context of this page.

        Uses the old linking code.  This is deprecated, use
        renderLinksIn or _preLink/_renderWithLinks instead.

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage().aq_parent.TestPage
        >>>
        >>> p._oldRenderLinksIn('http://a.b.c/d')
        '<a href="http://a.b.c/d">http://a.b.c/d</a>'
        >>> p._oldRenderLinksIn('mailto://a@b.c')
        '<a href="mailto://a@b.c">mailto://a@b.c</a>'
        >>> p._oldRenderLinksIn('nolink')
        'nolink'
        >>> p._oldRenderLinksIn('TestPage')[-23:]
        '/TestPage">TestPage</a>'
        >>> p._oldRenderLinksIn('TestPage1234')[-43:]
        '/TestPage/editform?page=TestPage1234">?</a>'
        >>> p._oldRenderLinksIn('!TestPage')
        'TestPage'
        >>> p._oldRenderLinksIn('[nolink]')[-37:]
        '/TestPage/editform?page=nolink">?</a>'
        >>> # a problem with escaping remote wiki links was reported
        >>> p.edit(text='RemoteWikiURL: URL/')
        >>> p._oldRenderLinksIn('TestPage:REMOTEPAGE') # should give:
        '<a href="URL/REMOTEPAGE">TestPage:REMOTEPAGE</a>'
        >>> p._oldRenderLinksIn('!TestPage:REMOTEPAGE') # should give:
        'TestPage:REMOTEPAGE'
        >>>
        >>> p._oldRenderLinksIn("[ ]")[-34:]
        '/TestPage/editform?page=%20">?</a>'

        >>> # do display the brackets prior to page creation
        >>> p._oldRenderLinksIn('[newpage]')[:9]
        '[newpage]'
        >>> # unconditionally link things that look like a path
        >>> p._oldRenderLinksIn('[folder/file]')
        '<a href="folder/file">folder/file</a>'
        """ #' 
        t = text
        if doLineEscapes:
            t = self.applyLineEscapesIn(t)
        t = re.sub(interwikilink,
                   thunk_substituter(self._replaceInterWikilink, t, 1),
                   t)
        t = re.sub(wikilink,
                   thunk_substituter(self._replaceWikilink, t, 1),
                   t)
        return t

    def _replaceInterWikilink(self, match, allowed=0, state=None, text='',
                              protect=1):
        """
        Replace an occurrence of interwikilink with a suitable hyperlink.

        To be used as a re.sub repl function *and* get a proper value
        for literal context, 'allowed', etc, enclose this function
        with the value using 'thunk_substituter'.

        Used by both _oldRenderLinksIn and _renderLink.
        """
        # matches beginning with ! should be left alone
        #if re.match('^!',match.group(0)): return match.group(1)
        # NB this is a bit naughty, but: since we know this text will likely
        # be scanned with _replaceWikilink right after this pass, leave
        # the ! in place for it to find. Otherwise the localname will
        # get wiki-linked.
        if re.match('^!',match.group(0)): return match.group(0)

        localname  = match.group('local')
        remotename = match.group('remote')

        # NB localname could be [bracketed]
        if re.match(bracketedexpr,localname):
            localname = re.sub(bracketedexpr, r'\1', localname)

        # look for a RemoteWikiURL definition
        if hasattr(self.folder(), localname): 
            localpage = getattr(self.folder(),localname)
            # local page found - search for "RemoteWikiUrl: url"
            m = re.search(remotewikiurl, str(localpage))
            if m is not None:
                remoteurl = html_unquote(m.group(1)) # NB: pages are 
                                                     # stored html-quoted
                                                     # XXX eh ? they are ?
                                                     # something's not
                                                     # right
                                                     # somewhere.. 
                                                     # I have lost my
                                                     # grip on this
                                                     # whole quoting
                                                     # issue.
                
                # we have a valid inter-wiki link
                link = '<a href="%s%s">%s:%s</a>' % \
                       (remoteurl, remotename, localname, remotename)
                if protect:
                    # normally, need to protect it from later wiki-izing passes
                    link = re.sub(wikilink, r'!\1', link)
                return link

        # otherwise, leave alone
        return match.group(0)

    def _replaceWikilink(self, match, allowed=0, state=None, text=''):
        """
        Replace an occurrence of the wikilink regexp or one of the
        special [] constructs with a suitable hyperlink

        To be used as a re.sub repl function *and* get a proper value
        for literal context, 'allowed', etc, enclose this function
        with the value using 'thunk_substituter'.

        Used by _oldRenderLinksIn; this lags behind _renderLink now.
        Deprecated.
        """
        # In a literal?
        if state is not None:
            if within_literal(match.start(1), match.end(1)-1, state, text):
                return match.group(1)

        # matches beginning with ! should be left alone
        if re.match('^!',match.group(0)):
            return match.group(1)

        m = morig = match.group(1)

        # if it's a bracketed expression,
        if re.match(bracketedexpr,m):

            # strip the enclosing []'s
            m = re.sub(bracketedexpr, r'\1', m)

            # extract a (non-url) path if there is one
            pathmatch = re.match(r'(([^/]*/)+)([^/]+)',m)
            if pathmatch:
                path,id = pathmatch.group(1), pathmatch.group(3)
            else:
                path,id = '',m

            # if it looks like an image link, inline it
            if guess_content_type(id)[0][0:5] == 'image':
                return '<img src="%s%s">' % (path,id)

            # or if there was a path assume it's to some non-wiki
            # object and skip the usual existence checking for
            # simplicity. Could also attempt to navigate the path in
            # zodb to learn more about the destination
            if path:
                #return '<a href="%s%s">%s</a>' % (path,id,id)
                return '<a href="%s%s">%s%s</a>' % (path,id,path,id)

            # otherwise fall through to normal link processing
        
        # if it's an ordinary url, link to it
        if re.match(url,m):
            # except, if preceded by " or = it should probably be left alone
            if re.match('^["=]',m):                                   # "
                return m
            else:
                return '<a href="%s">%s</a>' % (m, m)

        # it might be a structured text footnote ?
        # see _renderLink
        elif (morig[-1] == ']' and
              re.search(r'(?s)<a name="ref%s"' % (m),text)):
            return '<a href="#ref%s">[%s]</a>' % (m,m)

        # a wikiname - if a page (or something) of this name exists in
        # this folder, link to it
        elif hasattr(self.folder().aq_base, m): 
        #elif m in self.folder().objectIds():
            #return '<a href="%s">%s</a>' % (quote(m), m)
            # wiki links were made absolute for robustness in editform etc.
            # a relative_urls property can re-enable the old behaviour
            if getattr(self.folder(),'relative_urls',None):
                return '<a href="%s">%s</a>' % (quote(m), m)
            else:
                return '<a href="%s/%s">%s</a>' % (self.wiki_url(),quote(m), m)

        # if it's acquired from above, link it differently to facilitate
        # sub-wikis (one level)
        elif (hasattr(self.folder(),'aq_parent') and
              hasattr(self.folder().aq_parent, m)):
              #m in self.folder().aq_parent.objectIds()):
            if getattr(self.folder(),'relative_urls',None):
                return '<a href="../%s">../%s</a>' % (quote(m), m)
            else:
                return '<a href="%s/../%s">../%s</a>' % \
                       (self.wiki_url(),quote(m), m)

        # otherwise, provide a "?" creation link
        else:
            if getattr(self.folder(),'relative_urls',None):
                return '%s<a href="%s/editform?page=%s">?</a>' % \
                       (morig, quote(self.id()), quote(m))
            else:
                return '%s<a href="%s/%s/editform?page=%s">?</a>' % \
                       (morig, self.wiki_url(), quote(self.id()), quote(m))
        
    def _preLink(self):
        """
        find links in _prerendered & prepare for a fast _renderWithLinks

        saves a list of alternating text extents and link names in
        _prelinked and the unique link names in _links
        """
        get_transaction().note('prelink')
        text = self._prerendered or ''
        #XXX regexp alert - possibly slower than before with large pages
        r = anywikilinkexpr
        self._links = []
        self._prelinked = []
        lastpos = textstart = 0
        state = {'lastend':0,'inpre':0,'incode':0,'intag':0,'inanchor':0}
        while 1:
            m = r.search(text,lastpos)
            if m:
                link = m.group()
                linkstart,linkend = m.span()
                #if (not within_literal(m.start(1),m.end(1)-1,
                if (not within_literal(linkstart,linkend-1,
                                       state,text) and
                    #XXX more tests; easier than adapting within_literal
                    not withinSgmlOrDtml(m.span(),text) and
                    not re.match('^!',link)):
                    # found a link - save it and the text extent before it
                    self._links.append(link)
                    self._prelinked.append(text[textstart:linkstart])
                    self._prelinked.append(text[linkstart:linkend])
                    lastpos = textstart = linkend
                else:
                    # found the link pattern but it's escaped or inside a
                    # stx literal or sgml tag
                    # leave it as part of the ordinary text
                    # and strip the ! from escaped links
                    if re.match('^!',link):
                        text = text[:linkstart] + text[linkstart+1:]
                        linkend = linkend - 1
                    lastpos = linkend
            else:
                # no more links - save the final text extent & quit
                self._prelinked.append(text[textstart:])
                break

    def _renderWithLinks(self):
        """
        Regenerate this page's pre-rendered data with up-to-date links.

        For pre-linking page types, called on each page view (and also
        after setting text). Assumes _preLink has been called first.
        Renders all cached links and interpolates these with the static
        text extents previously identified.

        If the result is different from self._prerendered, save it back
        there. For dtml modes, re-cook in this case so dtml knows what's
        where. In most cases this will not be necessary.

        Tries to avoid excessive writes. Wiki links now include titles
        with relative last edit time and author and hence change a lot.
        To avoid zodb writes and possible re-cooking, we exclude these
        link titles from the prelinking process and fill them in at the
        last moment with _renderLateLinkTitles.
        """
        if not self._prelinked:
            return
        renderedlinks = self._renderedLinks()
        t = ''
        prelinked = self._prelinked
        for i in range(0,len(prelinked)-1,2):
            text,link = prelinked[i],prelinked[i+1]
            t = t + text
            t = t + renderedlinks[link]
        t = t + self._prelinked[-1]
        if t != self._prerendered:
            get_transaction().note('renderwithlinks')
            self._prerendered = t
            if self.supportsDtml():
                self.cook()

    def _renderedLinks(self):
        """
        return a mapping of our cached links to their current renderings
        """
        rendered = {}
        for link in self._links:
            rendered[link] = self._renderLink(link,render_title=0)
        return rendered

    def _renderLink(self,link,allowed=0,state=None,text='',render_title=1):
        """
        render a link appropriately depending on the current page/wiki context

        Can be called directly (link should be a string)
        or from re.sub via thunk substituter (link will be a match object).
        Based on _replaceWikilink & _replaceInterWikilink.

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> p._renderLink('[unbalanced (]')[-53:]
        'page=unbalanced%20%28" title="create this page">?</a>'
        """ #'
        if type(link) is not StringType:
            # we are being called from re.sub via thunk_substituter
            # do the within_literal and other checks that _preLink would
            # normally do
            match = link
            link = match.group()
            if (within_literal(match.start(),match.end()-1,state,text) or
                withinSgmlOrDtml(match.span(),text) or
                re.match('^!',link)):
                if re.match('^!',link):
                    link = link[1:]
                return link
        else:
            text = getattr(self,'_prerendered','') or ''

        folder = self.folder()
        m = re.match(interwikilink,link)
        if m:
            return self._replaceInterWikilink(m,protect=0)
        else:
            m = morig = link

        # if it's a bracketed phrase,
        if re.match(bracketedexpr,m):

            # strip the enclosing []'s
            m = re.sub(bracketedexpr, r'\1', m)

            # look for a page with a name matching this
            pagewiththisname = self.pageWithFuzzyName(m,ignore_case=1)
            if pagewiththisname:
                m = pagewiththisname.getId()
                # and fall through to normal link processing
                # XXX but m might get hassled by the next two if clauses

            else:

                # XXX old stuff - may restrict [] to wiki pages later

                # extract a (non-url) path if there is one
                # XXX could also be a [] link containing /
                # try it as a path & link if the thing exists, 
                # otherwise treat this as a [] link  ?
                # or leave out path linking altogether ?
                # yes, let's simplify []'s functionality for the moment
                #pathmatch = re.match(r'(([^/]*/)+)([^/]+)',m)
                #if pathmatch:
                #    path,id = pathmatch.group(1), pathmatch.group(3)
                #else:
                #    path,id = '',m

                # if it looks like an image link, inline it
                #if guess_content_type(id)[0][0:5] == 'image':
                #    return '<img src="%s%s">' % (path,id)

                # or if there was a path assume it's to some non-wiki
                # object and skip the usual existence checking for
                # simplicity. Could also attempt to navigate the path in
                # zodb to learn more about the destination
                #if path:
                #    #return '<a href="%s%s">%s</a>' % (path,id,id)
                #    return '<a href="%s%s">%s%s</a>' % (path,id,path,id)

                # otherwise fall through to normal link processing
                pass
        
        # if it's an ordinary url, link to it
        if re.match(url,m):
            # except, if preceded by " or = it should probably be left alone
            if re.match('^["=]',m):                                   # "
                return m
            else:
                return '<a href="%s">%s</a>' % (m, m)

        # it might be a structured text footnote ?
        # try to avoid matching accidentally
        # XXX needs to be in docs
        # XXX stx uses ref prefix now.. from what version ? what about older
        # zopes ?
        elif (morig[-1] == ']' and
              re.search(r'(?s)<a name="ref%s"' % (re.escape(m)),text)):
            return '<a href="#ref%s" title="footnote %s">[%s]</a>' % (m,m,m)

        # a wikiname - if a page of this name exists in
        # this folder, link to it
        #elif (hasattr(folder.aq_base, m) and
        #      self.isZwikiPage(getattr(folder.aq_base,m))):
        # XXX new: if international characters have been enabled for wiki
        # names but not page ids (see Regexps.py), wikinames too may
        # differ from their page id.  Are we sure we want to allow this ?
        # check quick case first.. premature optimization, all this will
        # be refactored
        elif self.pageWithNameOrId(m):
            #return '<a href="%s">%s</a>' % (quote(m), m)
            # wiki links were made absolute for robustness in editform etc.
            # a relative_urls property can re-enable the old behaviour
            morig = re.sub(r'\[(.*)\]',r'\1',morig)
            if render_title:
                linktitle = folder[m].linkTitle()
            else:
                linktitle = ''
            if getattr(folder,'relative_urls',None):
                return '<a href="%s" title="%s">%s</a>' % \
                       (quote(m),linktitle,morig)
            else:
                return '<a href="%s/%s" title="%s">%s</a>' % \
                       (self.wiki_url(),quote(m),linktitle,morig)

        # if it's acquired from above, link it differently to facilitate
        # sub-wikis (one level).
        # XXX NB pages whose wikiname and id differ as described above
        # will not be linked.
        elif (hasattr(folder,'aq_parent') and
              hasattr(folder.aq_parent, m) and
              self.isZwikiPage(getattr(folder.aq_parent,m))):
            if render_title:
                linktitle = folder[m].linkTitle()
            else:
                linktitle = ''
            if getattr(folder,'relative_urls',None):
                return \
                  '<a href="../%s" title="page in parent wiki, %s">../%s</a>' \
                  % (quote(m),linktitle,morig)
            else:
                return \
                  '<a href="%s/../%s" title="page in parent wiki, %s">../%s</a>'\
                  % (self.wiki_url(),quote(m),linktitle,morig)

        # otherwise, provide a "?" creation link
        else:
            if getattr(folder,'relative_urls',None):
                return '%s<a href="%s/editform?page=%s" title="create this page">?</a>' % \
                       (morig, quote(self.id()), quote(m))
            else:
                return '%s<a href="%s/%s/editform?page=%s" title="create this page">?</a>' % \
                       (morig, self.wiki_url(), quote(self.id()), quote(m))

    def renderLinksIn(self,text,doLineEscapes=0): #XXX default to 1 ?
        """
        Return text with all links rendered as for the context of this page.

        Like _oldRenderLinksIn but uses the newer linking code

        # tests based on _oldRenderLinksIn
        # clearer way to write these ?
        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage().aq_parent.TestPage
        >>>
        >>> p.renderLinksIn('nolink')
        'nolink'
        >>>
        >>> p.renderLinksIn('http://a.b.c/d')
        '<a href="http://a.b.c/d">http://a.b.c/d</a>'
        >>>
        >>> p.renderLinksIn('mailto://a@b.c')
        '<a href="mailto://a@b.c">mailto://a@b.c</a>'
        >>>
        >>> p.renderLinksIn('TestPage')
        '<a href="http://foo/foo/test//TestPage" title="last edited some time ago">TestPage</a>'
        >>>
        >>> p.renderLinksIn('NewTestPage')
        'NewTestPage<a href="http://foo/foo/test//TestPage/editform?page=NewTestPage" title="create this page">?</a>'
        >>>
        >>> p.renderLinksIn('!TestPage')
        'TestPage'
        >>>
        >>> p.renderLinksIn('[newpage]')
        '[newpage]<a href="http://foo/foo/test//TestPage/editform?page=newpage" title="create this page">?</a>'
        >>>
        >>> # a problem with escaping remote wiki links was reported
        >>> p.edit(text='RemoteWikiURL: URL/')
        >>> p.renderLinksIn('TestPage:REMOTEPAGE')
        '<a href="URL/REMOTEPAGE">TestPage:REMOTEPAGE</a>'
        >>> p.renderLinksIn('!TestPage:REMOTEPAGE')
        'TestPage:REMOTEPAGE'
        >>>
        >>> p.renderLinksIn("[ ]")
        '[ ]<a href="http://foo/foo/test//TestPage/editform?page=%20" title="create this page">?</a>'
        >>>
        >>> # do display the brackets prior to page creation
        >>> p.renderLinksIn('[newpage]')
        '[newpage]<a href="http://foo/foo/test//TestPage/editform?page=newpage" title="create this page">?</a>'
        >>>
        >>> # unconditionally link things that look like a path
        >>> #p.renderLinksIn('[folder/file]')
        >>> #'<a href="folder/file">folder/file</a>'
        >>>
        >>> # don't link wikinames inside <a href...>...</a>
        >>> p.renderLinksIn('<a href>WikiName</a>')
        '<a href>WikiName</a>'
        >>> # do link wikinames after <a name...> with no closing </a>
        >>> p.renderLinksIn('<a name>WikiName')
        '<a name>WikiName<a href="http://foo/foo/test//TestPage/editform?page=WikiName" title="create this page">?</a>'
        """
        
        t = text
        if doLineEscapes:
            t = self.applyLineEscapesIn(t)
        t = re.sub(anywikilinkexpr,
                   thunk_substituter(self._renderLink, t, 1),
                   t)
        return t

    def links(self):
        """
        list the links occurring on this page - useful for cataloging

        includes urls & interwiki links
        """
        # Make sure the _links cache is up to date.
        # It's ok to call _preLink even for pages which don't do
        # pre-linking. It expects pre-rendering to have been called first.
        if not self._prerendered:
            self._preRender()
        if not self._prelinked:
            self._preLink()
        return self._links

    def canonicalLinks(self):
        """
        list the canonical id form of the wiki links occurring on this page

        useful for calculating backlinks.
        """
        clinks = []
        localwikilinkexpr = re.compile(localwikilink)
        for link in self.links():
            if localwikilinkexpr.match(link):
                if link[0] == r'[' and link[-1] == r']':
                    link = link[1:-1]
                clink = self.canonicalIdFrom(link)
                clinks.append(clink)
        return clinks

    #def actualLinks(self):
    #    """
    #    list the actual page ids of local pages this page currently links to
    #
    #    These are dynamic, unlike links and canonicalLinks which are set
    #    at edit time.
    #    performance-sensitive
    #    """
    #    f = self.folder()
    #    canonicalids = map(lambda x:x.title_or_id(),
    #                       f.objectValues(spec=self.meta_type))
    #    for i in range(len(canonicalids)):
    #        canonicalids[i] = self.canonicalIdFrom(canonicalids[i])
    #    linkids = []
    #    localwikilinkexpr = re.compile(localwikilink)
    #    for link in self.links():
    #        if localwikilinkexpr.match(link):
    #            if link[0] == r'[' and link[-1] == r']':
    #                link = link[1:-1]
    #            clink = self.canonicalIdFrom(link)
    #            p = self.pageWithId(clink)
    #            if p:
    #                linkids.append(p.id())
    #            elif clink in canonicalids:
    #                p = self.pageWithFuzzyName(clink)
    #                # I expect p always to be found, but sometimes it isn't ?
    #                # eg tav, wavepowerarea
    #                # other problem pages for linkids:
    #                # UserOptionsDiscussion, ZWikiDiscussion200110,
    #                # WikiMailDiscussion, etc.
    #                if p:
    #                    linkids.append(p.id())
    #    return linkids

    ######################################################################
    # page naming methods

    def canonicalIdFrom(self,name):
        """
        convert a page name made up of arbitrary text to a canonical safe id

        Constraints for zwiki page ids:
        - it needs to be a legal zope object id
        - to simplify linking, we will require it to be a valid url
        - it should be unique for a given name (ignoring whitespace)
        - we'd like it to be as similar to the name and as simple to read
          and work with as possible
        - we'd like to encourage serendipitous linking between free-form
          and wikiname links & pages

        So this version
        - capitalizes and joins whitespace-separated words into a wikiname
        - converts any non-zope-and-url-safe characters and _ to _hexvalue
        - if the above results in an id beginning with _, prepends X
          (XXX this breaks the uniqueness requirement, better ideas ?)

        performance-sensitive
        
        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage()
        >>> p.canonicalIdFrom('WikiName')
        'WikiName'
        >>> p.canonicalIdFrom('ZWikiWikiNames2')
        'ZWikiWikiNames2'
        >>> p.canonicalIdFrom('a page with !')
        'APageWith_21'
        >>> p.canonicalIdFrom('a_page')
        'A_5fpage'
        >>> p.canonicalIdFrom('age')
        'X_c3Page'
        >>> p.canonicalIdFrom('_c3Page')
        'X_5fc3Page'
        """
        # capitalize whitespace-separated words (without losing existing
        # capitals) then strip whitespace
        id = ' '+name
        id = spaceandlowerexpr.sub(lambda m:string.upper(m.group(1)),id)
        id = string.join(string.split(id),'')
        # quote unsafe characters (see Regexps.py)
        safeid = ''
        for c in id:
            if zwikiidcharsexpr.match(c):
                safeid = safeid + c
            else:
                safeid = safeid + '_%02x' % ord(c)
        # ensure id doesn't begin with _
        if len(safeid) and safeid[0] == '_':
            safeid = 'X'+safeid
        return safeid

    def canonicalId(self):
        """
        give the canonical id of this page
        """
        return self.canonicalIdFrom(self.title_or_id())

    def pageWithId(self,id,url_quoted=0,ignore_case=0):
                           
        """
        Return the page which has this id, or None.

        Can also do a case-insensitive id search,
        and optionally unquote id.
        """
        if url_quoted:
            id = unquote(id)
        f = self.folder()
        p = getattr(f,id,None)
        if hasattr(f,id) and getattr(f,id).meta_type == self.meta_type:
            return getattr(f,id)
        elif ignore_case:
            id = string.lower(id)
            # use catalog ?
            for p in f.objectValues(spec=self.meta_type):
                if id == string.lower(p.id()):
                    return p
        else:
            return None

    def pageWithName(self,name,url_quoted=0):
        """
        return the page which has this name, or None

        page name may be different from page id, and if so is stored in
        the title property. Ie page name is currently defined as
        the value given by title_or_id().
        """
        if url_quoted:
            name = unquote(name)
        f = self.folder()
        # check easy case first
        if hasattr(f.aq_base, name):
            p = f[name]
            if self.isZwikiPage(p) and p.title_or_id() == name:
                return p
        # use catalog ?
        for p in f.objectValues(spec=self.meta_type):
            if p.title_or_id() == name:
                return p
        return None

    def pageWithNameOrId(self,name,url_quoted=0):
        """
        return the page which has this as it's name or id, or None

        see pageWithName. Are these the same ? not quite.
        """
        if url_quoted:
            name = unquote(name)
        f = self.folder()
        if hasattr(f.aq_base,name) and self.isZwikiPage(f[name]):
            return f[name]
        else:
            return self.pageWithName(name)
        
    def pageWithFuzzyName(self,name,url_quoted=0,
                           allow_partial=0,ignore_case=0):
        """
        return the page for which name is a valid fuzzy link, or None

        A fuzzy link converts all page names to their canonical id form
        before matching. allow_partial and ignore_case flags allow even
        fuzzier matching.

        We search for the common easy matches first for speed.
        After those, if there are multiple fuzzy matches return the page
        whose name is alphabetically first.

        NB can also match non-zwikipage objects at present

        performance-sensitive
        """
        if url_quoted:
            name = unquote(name)
        f = self.folder()
        p = self.pageWithNameOrId(name)
        if p: return p
        id = self.canonicalIdFrom(name)
        p = self.pageWithNameOrId(id)
        if p: return p
        if ignore_case:
            id = string.lower(id)
        p = self.pageWithNameOrId(id)
        if p: return p
        # that's all the easy cases.. now compare against the
        # canonical form of all existing page names
        # XXX issueno1 doesn't find IssueNo1 when ignore_case is true -
        # need to check all the ids as well
        # use catalog ?
        allnames = map(lambda x:x.title_or_id(),
                       f.objectValues(spec=self.meta_type))
        allnames.sort()
        for pname in allnames:
            pid = self.canonicalIdFrom(pname)
            if ignore_case:
                pid = string.lower(pid)
            if (pid == id or 
                (allow_partial and pid[:len(id)] == id)):
                return self.pageWithName(pname)
        return None
        
    ######################################################################
    # page editing/creation/deletion/file upload methods

    def _checkPermission(self, permission, object):
        return getSecurityManager().checkPermission(permission,object)

    #XXX title argument no longer valid, remove
    def create(self,page,text=None,type=None,title='',REQUEST=None,log=None):
        """
        Create a new wiki page and redirect there if appropriate; can
        upload a file at the same time.  Normally edit() will call
        this for you.

        Assumes page has been url-quoted. If it's not a url-safe name, we
        will create the page with a url-safe id that's similar. We assume
        this id won't match anything already existing (zwiki would have
        linked it instead of offering to create it).
        """
        # do we have permission ?
        if not self._checkPermission(Permissions.Add,self.folder()):
            raise 'Unauthorized', (
                _('You are not authorized to add ZWiki Pages here.'))

        name = unquote(page)
        id = self.canonicalIdFrom(name)

        # make a new (blank) page object, situate it
        # in the parent folder and get a hold of it's
        # normal acquisition wrapper
        # newid should be the same as id, but don't assume
        p = ZWikiPage(source_string='', __name__=id)
        newid = self.folder()._setObject(id,p)
        p = getattr(self.folder(),newid)

        p.title = name
        p._setCreator(REQUEST)
        p._setLastEditor(REQUEST)
        p._setLastLog(log)
        p._setOwnership(REQUEST)
        p.parents = [self.title_or_id()]

        # inherit type from the previous (parent) page if not specified
        # or overridden via folder attribute,
        # special case: don't inherit issuedtml page type
        # XXX standard_page_type prevents creating issue pages.
        # What to do ? For now nothing, assume the tracker add form
        # has permission to change the page type afterward
        if hasattr(self.folder(),'standard_page_type'):
            p.page_type = self.folder().standard_page_type
        elif type:
            p.page_type = type
        elif self.page_type != 'issuedtml':
            p.page_type = self.page_type
        else:
            p.page_type = DEFAULT_PAGE_TYPE

        # set initial page text as edit() would, with cleanups and dtml
        # validation
        p._setText(text or '',REQUEST)

        # if a file was submitted as well, handle that
        p._handleFileUpload(REQUEST)

        # we got indexed after _setObject,
        # but do it again with our text in place
        p.index_object()
        
        if p.usingRegulations():
            # initialize regulations settings.
            p.setRegulations(REQUEST,new=1)

        # redirect browser if needed
        if REQUEST is not None:
            try:
                u=REQUEST['URL2']
                REQUEST.RESPONSE.redirect(u + '/' + quote(newid))
            except KeyError:
                pass

    def comment(self, text='', username='', time='', note=' (via web)',
                use_heading=0, REQUEST=None):
        """
        A handy method, like append but

        - adds a standard comment heading with user name, time & note

        - allows those fields to be specified for some flexibility
        with mail-ins etc

        - does some html prettification for web display; nb other
        parts try to strip this same html (sendMailSubscriber)

        As a convenience for standard_wiki_footer, unless use_heading
        is true we act exactly like append. Should think of something
        better here.

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage().aq_parent.TestPage
        >>> p.edit(text='test')
        >>> p.comment(text='comment',username='me',time='now')
        >>> assert p.read() == 'test\\n\\ncomment'
        >>>
        >>> p.comment(text='comment',username='me',time='now',use_heading=1)
        >>> assert p.read() == 'test\\n\\ncomment\\n\\n<hr><b>me, now (via web):</b><br>\\ncomment'

        the \\n's above are \012's with zope < 2.4
        """
        oldtext = self.read()

        if not use_heading:
            self.append(text=text,REQUEST=REQUEST)

        else:
            # generate the comment heading
            if not username:
                username = self.zwiki_username_or_ip(REQUEST)
            # don't show ip addresses
            if re.match(r'^[0-9\.\s]*$',username): 
                username = ''
            separator = username and ', '
            if not time:
                time = self.ZopeTime().strftime('%Y/%m/%d %H:%M %Z')
            #XXX temp hack - don't use rule in CMF
            if self.inCMF():
                heading = '\n\n<b>%s%s%s%s:</b><br>\n' % \
                          (html_quote(username),
                           separator,
                           html_quote(time),
                           html_quote(note))
            else:
                heading = '\n\n<hr><b>%s%s%s%s:</b><br>\n' % \
                          (html_quote(username),
                           separator,
                           html_quote(time),
                           html_quote(note))
    
            # italicize quoted text in replies
            text = re.sub(r'(?m)^>(.*)',r'<br>><i>\1</i>',text)
    
            self.append(text,separator=heading,REQUEST=REQUEST)

        # if policy is comments only or none, do mailout here
        if ((not hasattr(self.folder(),'mailout_policy')
             or self.folder().mailout_policy == 'comments') and
            text):
            # get the username in there
            if REQUEST: REQUEST.cookies['zwiki_username'] = username
            self.sendMailToSubscribers(
                self.textDiff(a=oldtext,b=self.read(),verbose=0),
                REQUEST,
                "")

    def append(self, text='', separator='\n\n', REQUEST=None):
        """
        Appends some text to an existing zwiki page by calling edit;
        may result in mail notifications to subscribers.
        """
        oldtext = self.read()
        text = str(text)
        if text:
            # usability hack: scroll to bottom after adding a comment
            if REQUEST:
                REQUEST['URL1'] = REQUEST['URL1'] + '#bottom'
            self.edit(text=oldtext+separator+text, REQUEST=REQUEST)

    def edit(self, page=None, text=None, type=None, title='', 
             timeStamp=None, REQUEST=None, 
             subjectSuffix='', log=None, check_conflict=1): # temp
        """
        General-purpose method for editing & creating zwiki pages.
        Changes the text and/or markup type of this (or the specified)
        page, or creates the specified page (name or id allowed) if it
        does not exist.
        
        Other special features:

        - Usually called from a time-stamped web form; we use
        timeStamp to detect and warn when two people attempt to work
        on a page at the same time. This makes sense only if timeStamp
        came from an editform for the page we are actually changing.

        - The username (authenticated user or zwiki_username cookie)
        and ip address are saved in page's last_editor, last_editor_ip
        attributes if a change is made

        - If the text begins with "DeleteMe", move this page to the
        recycle_bin subfolder.

        - If file has been submitted in REQUEST, create a file or
        image object and link or inline it on the current page.

        - May also cause mail notifications to be sent to subscribers

        This code has become more complex to support late page
        creation, but the api should now be more general & powerful
        than it was.  Doing all this stuff in one method simplifies
        the layer above I think.
        """
        
        #self._validateProxy(REQUEST)   # XXX correct ? don't think so
                                        # do zwiki pages obey proxy roles ?

        if page: page = unquote(page)
        # are we changing this page ?
        if page is None:
            p = self
        # changing another page ?
        elif self.pageWithNameOrId(page):
            p = self.pageWithNameOrId(page)
        # or creating a new page
        else:
            return self.create(page,text,type,title,REQUEST,log)

        # ok, changing p. We may be doing several things here;
        # each of these handlers checks permissions and does the
        # necessary. Some of these can halt further processing.
        # todo: tie these in to mail notification, along with 
        # other changes like reparenting
        if check_conflict and self.checkEditConflict(timeStamp, REQUEST):
            return self.editConflictDialog()
        if check_conflict and hasattr(self,'wl_isLocked') and self.wl_isLocked():
            return self.davLockDialog()
        if p._handleDeleteMe(text,REQUEST,log):
            return
        p._handleEditPageType(type,REQUEST,log)
        p._handleEditText(text,REQUEST,subjectSuffix,log)
        p._handleFileUpload(REQUEST,log)
        if self.usingRegulations():
            p._handleSetRegulations(REQUEST)

        # update catalog if present
        try:
            p.index_object()
        except:
            # XXX show traceback in log
            DLOG('failed to index '+p.id())
            try:
                p.reindex_object()
            except:
                DLOG('failed to reindex '+p.id()+', giving up')
                pass

        # redirect browser if needed
        if REQUEST is not None:
            u=REQUEST.get('redirectURL',REQUEST['URL1'])
            REQUEST.RESPONSE.redirect(u)

    def _handleSetRegulations(self,REQUEST):
        # do we have permission ?
        if not self._checkPermission(Permissions.ChangeRegs,self):
            raise 'Unauthorized', (
              _("You are not authorized to set this ZWiki Page's regulations."))
        self.setRegulations(REQUEST)
        self._preRender(clear_cache=1)
        #self._setLastEditor(REQUEST)

    def _handleEditPageType(self,type,REQUEST=None,log=None):
        # is the new page type valid and different ?
        if (type is not None and
            type != self.page_type):
            # do we have permission ?
            if not self._checkPermission(Permissions.ChangeType,self):
                raise 'Unauthorized', (
                    _("You are not authorized to change this ZWiki Page's type."))
            # change it
            self.page_type = type
            self._preRender(clear_cache=1)
            self._setLastEditor(REQUEST)
            self._setLastLog(log)

    def _setLastLog(self,log):
        """\
        Note log message, if provided.
        """
        if log and string.strip(log):
            log = string.strip(log)
            get_transaction().note('"%s"' % log)
            self.last_log = log
        else:
            self.last_log = ''


    def _handleEditText(self,text,REQUEST=None, subjectSuffix='', log=None):
        # is the new text valid and different ?
        if (text is not None and
            self._cleanupText(text) != self.read()):
            # do we have permission ?
            if (not
                (self._checkPermission(Permissions.Change, self) or
                 (self._checkPermission(Permissions.Append, self)
                  and find(self._cleanupText(text),self.read()) == 0))):
                raise 'Unauthorized', (
                    _('You are not authorized to edit this ZWiki Page.'))

            # change it
            oldtext = self.read()
            self._setText(text,REQUEST)
            self._setLastEditor(REQUEST)
            self._setLastLog(log)

            # if mailout policy is "all edits", do it here
            if (hasattr(self.folder(),'mailout_policy')
                and self.folder().mailout_policy == 'edits'):
                self.sendMailToSubscribers(
                    self.textDiff(a=oldtext,b=self.read()),
                    REQUEST,
                    "")

    def _handleDeleteMe(self,text,REQUEST=None,log=None):
        if not text or not re.match('(?m)^DeleteMe', text):
            return 0
        if (not
            (self._checkPermission(Permissions.Change, self) or
             (self._checkPermission(Permissions.Append, self)
              and find(self._cleanupText(text),self.read()) == 0))):
            raise 'Unauthorized', (
                _('You are not authorized to edit this ZWiki Page.'))
        if not self._checkPermission(Permissions.Delete, self):
            raise 'Unauthorized', (
                _('You are not authorized to delete this ZWiki Page.'))
        self._setLastLog(log)
        self._recycle(REQUEST)

        if REQUEST:
            # redirect to first existing parent, or front page
            destpage = ''
            for p in self.parents:
                if hasattr(self.folder(),p):
                    destpage = p
                    break
            REQUEST.RESPONSE.redirect(self.wiki_url()+'/'+quote(destpage))
            # I used to think redirect did not return, guess I was wrong

        # return true to terminate edit processing
        return 1

    def delete(self,REQUEST=None):
        """
        delete (recycle) this page, if permissions allow
        """
        if not (self._checkPermission(Permissions.Delete, self) and
                self.requestHasSomeId(REQUEST)):
            raise 'Unauthorized', (
                _('You are not authorized to delete this ZWiki Page.'))
        self._recycle(REQUEST)
        if REQUEST:
            # redirect to first existing parent or front page
            destpage = ''
            for p in self.parents:
                if hasattr(self.folder(),p):
                    destpage = p
                    break
            REQUEST.RESPONSE.redirect(self.wiki_url()+'/'+quote(destpage))

    def _recycle(self, REQUEST=None):
        """
        move this page to the recycle_bin subfolder, creating it if necessary.
        """
        # create recycle_bin folder if needed
        f = self.folder()
        if not hasattr(f,'recycle_bin'):
            f.manage_addFolder('recycle_bin', 'deleted wiki pages')
        # & move page there
        id=self.id()
        cb=f.manage_cutObjects(id)

        # update catalog if present
        # don't let manage_afterAdd catalog the new location..
        save = ZWikiPage.manage_afterAdd # not self!
        ZWikiPage.manage_afterAdd = lambda self,item,container: None
        f.recycle_bin.manage_pasteObjects(cb)
        ZWikiPage.manage_afterAdd = save

    def rename(self,pagename,leaveplaceholder=0,updatebacklinks=0,
               REQUEST=None):
        """
        Rename this page, if permissions allow, with optional wiki fixups.

        The options are

        - leave a placeholder page

        - update links throughout the wiki. Warning, this replaces all
        occurrences of the old page name (beginning and ending with a word
        boundary), not just links. This tries to do the most sensible
        thing when dealing with freeform page names, but probably won't
        find all possible fuzzy links.
        """
        # rename permission is also declared above.. all due for cleanup
        if not (self._checkPermission(Permissions.Rename, self) and
                self.requestHasSomeId(REQUEST)):
            raise 'Unauthorized', (
                _('You are not authorized to rename this ZWiki Page.'))
        newname = pagename
        oldid, oldname = self.id(), self.title_or_id()
        if newname and newname != oldname:
            self.title = newname
            newid = self.canonicalIdFrom(newname)
            if newid != oldid:
                self.folder().manage_renameObject(oldid,newid,REQUEST)
                self.index_object()
                if leaveplaceholder:
                    if newname == newid:
                        link = newname
                    else:
                        link = '[' + newname + ']'
                    self.create(oldid,
                                _("This page was renamed to %s.\n") % (link))
            if updatebacklinks:
                for p in self.backlinksFor(oldname):
                    oldtext = p.read()
                    newtext = re.sub(r'\b%s\b'%(oldname),newname,oldtext)
                    p.edit(text=newtext,REQUEST=REQUEST)
                if newid != oldid:
                    for p in self.backlinksFor(oldid):
                        oldtext = p.read()
                        newtext = re.sub(r'\b%s\b'%(oldid),newid,oldtext)
                        p.edit(text=newtext,REQUEST=REQUEST)
                    
            if REQUEST:
                REQUEST.RESPONSE.redirect(self.wiki_url()+'/'+quote(newid))

    def _handleFileUpload(self,REQUEST,log=None):
        # is there a file upload ?
        if (REQUEST and
            hasattr(REQUEST,'file') and
            hasattr(REQUEST.file,'filename') and
            REQUEST.file.filename):     # XXX do something

            # figure out the upload destination
            if hasattr(self,'uploads'):
                uploaddir = self.uploads
            else:
                uploaddir = self.folder()

            # do we have permission ?
            if not (self._checkPermission(Permissions.Upload,uploaddir)):# or
                    #self._checkPermission(Permissions.UploadSmallFiles,
                    #                self.folder())):
                raise 'Unauthorized', (
                    _('You are not authorized to upload files here.'))
            if not (self._checkPermission(Permissions.Change, self) or
                    self._checkPermission(Permissions.Append, self)):
                raise 'Unauthorized', (
                    _('You are not authorized to add a link on this ZWiki Page.'))
            # can we check file's size ?
            # yes! len(REQUEST.file.read()), apparently
            #if (len(REQUEST.file.read()) > LARGE_FILE_SIZE and
            #    not self._checkPermission(Permissions.Upload,
            #                        uploaddir)):
            #    raise 'Unauthorized', (
            #        _("""You are not authorized to add files larger than
            #        %s here.""" % (LARGE_FILE_SIZE)))

            # create the File or Image object
            file_id, content_type, size = \
                    self._createFileOrImage(REQUEST.file,
                                            title=REQUEST.get('title', ''),
                                            REQUEST=REQUEST)
            if file_id:
                # link it on the page and finish up
                self._addFileLink(file_id, content_type, size, REQUEST)
                self._setLastLog(log)
                self.index_object()
            else:
                # failed to create - give up (what about an error)
                pass

    def _createFileOrImage(self,file,title='',REQUEST=None,parent=None):
        # based on WikiForNow which was based on
        # OFS/Image.py:File:manage_addFile
        """
        Add a new File or Image object, depending on file's filename
        suffix. Returns a tuple containing the new id, content type &
        size, or (None,None,None).
        """
        # set id & title from filename
        title=str(title)
        id, title = OFS.Image.cookId('', title, file)
        if not id:
            return None, None, None

        # find out where to store files - in the 'uploads'
        # folder if defined, otherwise the wiki folder
        # NB a page might override this with a property
        if hasattr(self,'uploads'):
            folder = self.uploads
        else:
            folder = parent or self.folder() # see create()

        if hasattr(folder,id) and folder[id].meta_type in ('File','Image'):
            pass
        else:
            # First, we create the file or image object without data
            if guess_content_type(file.filename)[0][0:5] == 'image':
                folder._setObject(id, OFS.Image.Image(id,title,''))
            else:
                folder._setObject(id, OFS.Image.File(id,title,''))

        # Now we "upload" the data.  By doing this in two steps, we
        # can use a database trick to make the upload more efficient.
        folder._getOb(id).manage_upload(file)

        return id, folder._getOb(id).content_type, folder._getOb(id).getSize()

    def _addFileLink(self, file_id, content_type, size, REQUEST):
        """
        Add a link to the specified file at the end of this page,
        unless a link already exists.
        If the file is an image and not too big, inline it instead.
        """
        if re.search(r'(src|href)="%s"' % file_id,self.text()): return

        if hasattr(self,'uploads'):
            filepath = 'uploads/'
        else:
            filepath = ''
        if content_type[0:5] == 'image' and \
           not (hasattr(REQUEST,'dontinline') and REQUEST.dontinline) and \
           size <= LARGE_FILE_SIZE :
            linktxt = '\n\n<img src="%s%s">\n' % (filepath,file_id)
        else:
            linktxt = '\n\n<a href="%s%s">%s</a>\n' % (filepath,file_id,file_id)
        self._setText(self.read()+linktxt,REQUEST)
        self._setLastEditor(REQUEST)

    def _setOwnership(self, REQUEST=None):
        """
        set up the zope ownership of a new page appropriately

        depends on whether we are using regulations or not
        """
        if not self.usingRegulations():
            # To help control executable content, make sure the new page
            # acquires it's owner from the parent folder.
            self._deleteOwnershipAfterAdd()
        else:
            self._setOwnerRole(REQUEST)
            
    def _setText(self, text='', REQUEST=None):
        """
        Change the page text, with cleanups and perhaps DTML validation.

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage().aq_parent.TestPage
        >>> p.edit(text='<html blah>\\n<body blah>\\ntest\\n</body>\\n</html>')
        >>> p.read()
        '\\ntest\\n'
        
        """
        self.raw = self._cleanupText(text)
        self._preRender(clear_cache=1)
        if self.supportsPreLinking():
            self._renderWithLinks()
        # XXX here we should do DTML validation to catch & roll back
        # errors, but were getting authorization problems
        #if self.supportsDtml():
            # a commit didn't seem to help
            #get_transaction().commit()
            # so we skipped it for the moment
            #DTMLDocument.__call__(self,self,REQUEST,REQUEST.RESPONSE)

    def _cleanupText(self, t):
        """do some cleanup of a page's new text
        """
        # strip any browser-appended ^M's
        t = re.sub('\r\n', '\n', t)

        # convert international characters to HTML entities for safekeeping
        #for c,e in intl_char_entities:
        #    t = re.sub(c, e, t)
        # assume today's browsers will not harm these.. if this turns out
        # to be false, do some smarter checking here

        # here's the place to strip out any disallowed html/scripting elements
        # XXX there are updates for this somewhere on zwiki.org
        if DISABLE_JAVASCRIPT:
            t = re.sub(javascriptexpr,r'&lt;disabled \1&gt;',t)

        # strip out HTML document header/footer if added
        t = re.sub(htmlheaderexpr,'',t)
        t = re.sub(htmlfooterexpr,'',t)

        return t

    def _setLastEditor(self, REQUEST=None):
        """
        record my last_editor & last_editor_ip

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> from Products.ZWiki.tests.support import MockUser
        >>> p = MockZWikiPage().aq_parent.TestPage
        >>> p.REQUEST.set('REMOTE_ADDR', '1.2.3.4')
        >>> p.REQUEST.set('AUTHENTICATED_USER', MockUser('user'))
        >>> p._setLastEditor(REQUEST=p.REQUEST)
        >>> assert p.last_editor == 'user'
        >>> assert p.last_editor_ip == '1.2.3.4'
        """
        if REQUEST:
            self.last_editor_ip = REQUEST.REMOTE_ADDR
            self.last_editor = self.zwiki_username_or_ip(REQUEST)
        else:
            # this has been fiddled with before
            # if we have no REQUEST, at least update last editor
            self.last_editor_ip = ''
            self.last_editor = ''
        self.last_edit_time = DateTime(time.time()).ISO()

    def _setCreator(self, REQUEST=None):
        """
        record my creator, creator_ip & creation_time

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> from Products.ZWiki.tests.support import MockUser, MockRequest
        >>> p = MockZWikiPage().aq_parent.TestPage
        >>> r = MockRequest()
        >>> u = MockUser('test user')
        >>> r.set('AUTHENTICATED_USER',u)
        >>> r.set('REMOTE_ADDR','4.3.2.1')
        >>> 
        >>> p._setCreator()
        >>> assert p.creation_time
        >>> assert p.creator_ip == ''
        >>> assert p.creator == ''
        >>> 
        >>> p._setCreator(r)
        >>> assert p.creator_ip == '4.3.2.1'
        >>> assert p.creator == 'test user'

        """
        self.creation_time = DateTime(time.time()).ISO()
        if REQUEST:
            self.creator_ip = REQUEST.REMOTE_ADDR
            self.creator = self.zwiki_username_or_ip(REQUEST)
        else:
            self.creator_ip = ''
            self.creator = ''

    def checkEditConflict(self, timeStamp, REQUEST):
        """
        Warn if this edit would be in conflict with another.

        Edit conflict checking based on timestamps -
        
        things to consider: what if
        - we are behind a proxy so all ip's are the same ?
        - several people use the same cookie-based username ?
        - people use the same cookie-name as an existing member name ?
        - no-one is using usernames ?

        strategies:
        0. no conflict checking

        1. strict - require a matching timestamp. Safest but obstructs a
        user trying to backtrack & re-edit. This was the behaviour of
        early zwiki versions.

        2. semi-careful - record username & ip address with the timestamp,
        require a matching timestamp or matching non-anonymous username
        and ip.  There will be no conflict checking amongst users with the
        same username (authenticated or cookie) connecting via proxy.
        Anonymous users will experience strict checking until they
        configure a username.

        3. relaxed - require a matching timestamp or a matching, possibly
        anonymous, username and ip. There will be no conflict checking
        amongst anonymous users connecting via proxy. This is the current
        behaviour.
        """
        username = self.zwiki_username_or_ip()
        if (timeStamp is not None and
            timeStamp != self.timeStamp() and
            (not hasattr(self,'last_editor') or
             not hasattr(self,'last_editor_ip') or
             username != self.last_editor or
             REQUEST.REMOTE_ADDR != self.last_editor_ip)):
            return 1
        else:
            return 0

    def timeStamp(self):
        return str(self._p_mtime)
    
    def manage_FTPget(self):
        """Get source for FTP download.
        """
        #candidates = list(self.ZWIKI_PAGE_TYPES)
        #types = "%s (alternatives:" % self.page_type
        #if self.page_type in candidates:
        #    candidates.remove(self.page_type)
        #for i in candidates:
        #    types = types + " %s" % i
        #types = types + ")"
        types = "%s" % self.page_type
        return "Wiki-Safetybelt: %s\nType: %s\nLog: \n\n%s" % (
            self.timeStamp(), types, self.read())

    def old_PUT(self, REQUEST, RESPONSE):
        """Handle HTTP/FTP/WebDav PUT requests."""
        self.dav__init(REQUEST, RESPONSE)
        body=REQUEST.get('BODY', '')
        self._validateProxy(REQUEST)
        m=re.match(r'Wiki-Safetybelt: ([0-9]+[.][0-9]+)[\n][\n]?', body)
        if m:
            body = body[m.span()[1]:]            
            if (self.checkEditConflict(m.group(1), REQUEST) or
                (hasattr(self,'wl_isLocked') and self.wl_isLocked())):
                RESPONSE.setStatus(404) #XXX what should this be
                return RESPONSE
        self._setText(body)
        self._setLastEditor(REQUEST)
        self.index_object()
        RESPONSE.setStatus(204)
        return RESPONSE

    def PUT(self, REQUEST, RESPONSE):
        """Handle HTTP/FTP/WebDav PUT requests."""
        self.dav__init(REQUEST, RESPONSE)
        self.dav__simpleifhandler(REQUEST, RESPONSE, refresh=1)
        body=REQUEST.get('BODY', '')
        self._validateProxy(REQUEST)

        headers, body = parseHeadersBody(body)
        log = string.strip(headers.get('Log', headers.get('log', ''))) or None
        type = (string.strip(headers.get('Type', headers.get('type', '')))
                or None)
        if type is not None:
            type = string.split(type)[0]
            if type not in self.ZWIKI_PAGE_TYPES[:]:
                # Silently ignore it.
                type = None
        timestamp = string.strip(headers.get('Wiki-Safetybelt', '')) or None
        if timestamp and self.checkEditConflict(timestamp, REQUEST):
            RESPONSE.setStatus(423) # Resource Locked
            return RESPONSE

        #self._setText(body)
        #self._setLastEditor(REQUEST)
        #self.index_object()
        #RESPONSE.setStatus(204)
        #return RESPONSE
        try:
            self.edit(text=body, type=type, timeStamp=timestamp,
                      REQUEST=REQUEST, log=log, check_conflict=0)
        except 'Unauthorized':
            RESPONSE.setStatus(401)
            return RESPONSE
        RESPONSE.setStatus(204)
        return RESPONSE

    ######################################################################
    # UI methods
    #
    # methods to generate standard forms & dialogs; most of these can be
    # overridden by a DTML method of the same name. 
    # Return a placeholder if there is an error; useful if someone's
    # custom header has broken or if we are running tests.

    # make this a page method so tests/support.py can override it
    # and perhaps flesh it out later
    wikipage = default_wikipage

    def standard_wiki_header(self,client=None,REQUEST=None,RESPONSE=None,**kw):
        """
        Return a default or custom wiki page header. 

        May be overridden by a DTML method of the same name.
        """
        try:
            if hasattr(self.folder(), 'standard_wiki_header'):
                return self.folder().standard_wiki_header(self,REQUEST)
            else:
                return default_standard_wiki_header(self,self,REQUEST)
        except:
            type, val, tb = sys.exc_info()
            err = string.join(traceback.format_exception(type,val,tb),'')
            return '<html>\n<body>\ncould not render this.\n<!--\n%s\n-->\n'\
                   % err

    def standard_wiki_footer(self,client=None,REQUEST=None,RESPONSE=None,**kw):
        """
        Return a default or custom wiki page footer. 

        May be overridden by a DTML method of the same name.
        """
        try:
            if hasattr(self.folder(), 'standard_wiki_footer'):
                return self.folder().standard_wiki_footer(self,REQUEST)
            else:
                return default_standard_wiki_footer(self,self,REQUEST)
        except:
            type, val, tb = sys.exc_info()
            err = string.join(traceback.format_exception(type,val,tb),'')
            return 'could not render this.\n<!--\n%s\n-->\n</body>\n</html>\n'\
                   % err

    def backlinks(self, REQUEST=None):
        """
        Display a default or custom backlinks page. 

        May be overridden by a page template or DTML method of the same name.
        """
        folder = self.folder()
        if hasattr(folder,'backlinks'):
            if folder.backlinks.meta_type == 'Page Template':
                return folder.backlinks(this=self)
                #return folder.backlinks.__of__(self)() ?
            else:
                return folder.backlinks(self,REQUEST)
        else:
            return default_backlinks(self,self,REQUEST)

    def editform(self, REQUEST=None, page=None, text=None, action='ѹ'):
        """
        Display a default or custom form to edit or create the specified page.

        For new pages, initial text may be specified.
        May be overridden by a page template or DTML method of the same name.
        """
        if hasattr(self,'wl_isLocked') and self.wl_isLocked():
            return self.davLockDialog()

        # what are we going to do ? set up page, text & action accordingly
        if page is None:
            # no page specified - editing the current page
            page = self.title_or_id()
            text = self.read()
        #elif hasattr(self.folder(), page):
        elif self.pageWithName(page):
            # editing a different page
            text = self.pageWithName(page).read()
        else:
            # editing a brand-new page
            action = ''
            text = ''

        # display the edit form - a dtml method or the builtin default
        # NB we redefine id as a convenience, so that one header can work
        # for pages and editforms
        # XXX can we simplify this/make dtml more version independent ?
        # no way to restrict /editform currently
        folder = self.folder()
        if hasattr(folder,'editform'):
            if folder.editform.meta_type == 'Page Template':
                return folder.editform(this=self,page=page,text=text,
                                       action=action,id=page,oldid=self.id())
            else:
                return folder.editform(self,REQUEST,
                                       page=page,text=text,action=action,
                                       id=page,oldid=self.id())
        else:
            return default_editform(self,self,REQUEST,
                                    page=page,text=text,action=action,
                                    id=page,oldid=self.id())

    def subscribeform(self, REQUEST=None):
        """
        Display a default or custom mail subscription form. 

        May be overridden by a page template or DTML method of the same name.
        """
        folder = self.folder()
        if hasattr(folder,'subscribeform'):
            if folder.subscribeform.meta_type == 'Page Template':
                return folder.subscribeform(this=self)
            else:
                return folder.subscribeform(self,REQUEST)
        else:
            return default_subscribeform(self,self,REQUEST)
    
    def editConflictDialog(self):
        """
        web page displayed in edit conflict situations.
        """
        titlestr=_('Edit conflict')
        return MessageDialog(
            title=titlestr,
            message="""
            <b>%s</b>
            <p>
            %s.
            %s:
            <ol>
            <li>%s
            <li>%s
            <li>%s
            <li>%s
            <li>%s.
            </ol>
            %s,
            <p>
            %s.
            """ % (
            titlestr,
            _("Someone else has saved this page while you were editing"),
            _("To resolve the conflict, do this"),
            _("Click your browser's back button"),
            _("Copy your recent edits to the clipboard"),
            _("Click your browser's refresh button"),
            _("Paste in your edits again, being mindful of the latest changes"),
            _("Click the Change button again"),
            _("or"),
            _("To discard your changes and start again, click OK"),
            ),
            action=self.page_url()+'/editform')

    def davLockDialog(self):
        """
        web page displayed in webDAV lock conflict situations.
        """
        titlestr=_('Page is locked')
        return MessageDialog(
            title=titlestr,
            message="""
            <b>%s</b>
            <p>
            %s
            <p>
            %s
            """ % (
            titlestr,
            _("""
            This page has a webDAV lock. Someone is probably editing it
            with an external editor.  You'll need to wait until they've
            finished and then try again.  If you've just made some changes,
            you may want to back up and copy your version of the text for
            reference.
            """),
            _("To discard your changes and try again, click OK."),
            ),
            action=self.page_url()+'/editform')

    ######################################################################
    # misc methods

    def isZwikiPage(self,object):
        return getattr(object,'meta_type',None) == self.meta_type

    def backlinksFor(self, page):
        """
        Return the backlinks for page as a list of page objects

        (in no particular order).
        Currently used by rename, could be used by the backlinks forms too.

        XXX doesn't do fuzzy backlinks like the latest backlinks page,
        should.
        """
        regex = re.compile(r'\b%s\b'%(page))
        list = []
        # use catalog ?
        for p in self.folder().objectValues(spec=self.meta_type):
            if regex.search(p.read()):
                list.append(p)
        return list

    def zwiki_version(self):
        """
        Return the zwiki product version.
        """
        return __version__

    # expose these darn things for dtml programmers once and for all!
    # XXX safe ?
    # how are these public without docstrings ?
    def htmlquote(self, text):
        return html_quote(text)

    def htmlunquote(self, text):
        return html_unquote(text)

    def urlquote(self, text):
        return quote(text)

    def urlunquote(self, text):
        return unquote(text)

    zwiki_username_or_ip__roles__ = None
    def zwiki_username_or_ip(self, REQUEST=None):
        """
        search REQUEST for an authenticated member or a zwiki_username
        cookie or a username passed in by mailin.py

        XXX added REQUEST arg at one point when sending mail
        notification in append() was troublesome - still needed ?
        refactor
        """
        username = None
        REQUEST = REQUEST or getattr(self,'REQUEST',None)
        if REQUEST:
            username = REQUEST.get('MAILIN_USERNAME',None)
            if not username:
                user = REQUEST.get('AUTHENTICATED_USER')
                if user:
                    username = user.getUserName()
                if not username or username == str(user.acl_users._nobody):
                    username = REQUEST.cookies.get('zwiki_username',None)
                    if not username:
                        username = REQUEST.REMOTE_ADDR
        return username or ''

    def requestHasSomeId(self,REQUEST):
        """
        Check REQUEST has either a non-anonymous user or a username cookie.
        """
        username = self.zwiki_username_or_ip(REQUEST)
        if (username and username != REQUEST.REMOTE_ADDR):
            return 1
        else:
            return 0

    text__roles__ = None
    def text(self, REQUEST=None, RESPONSE=None):
        # see also backwards compatibility section
        # why permission-free ?
        """
        return this page's raw text
        (a permission-free version of document_src)
        also fiddle the mime type for web browsing

        misc tests:

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage().aq_parent.TestPage
        >>> 
        >>> # ensure we don't lose first lines to DTML's decapitate()
        >>> p._setText(r'first: line\\n\\nsecond line\\n')
        >>> p.text()
        'first: line\\\\n\\\\nsecond line\\\\n'
        >>> p.text()
        'first: line\\\\n\\\\nsecond line\\\\n'
        >>>
        >>> # ensure none of these reveal the antidecapkludge
        >>> p.edit(type='htmldtml')
        >>> p._setText('test text')
        >>> p.text()
        'test text'
        >>> from Products.ZWiki.ZWikiPage import ZOPEVERSION
        >>> if ZOPEVERSION >= (2,4):
        ...     assert p.read() == 'test text'
        ...     assert p.__str__() == 'test text'
        >>> 

        """
        if RESPONSE is not None:
            RESPONSE.setHeader('Content-Type', 'text/plain')
            #RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime))
        return self.read()

    # cf _setText, IssueNo0157
    _old_read = DTMLDocument.read
    def read(self):
        return re.sub('<!--antidecapitationkludge-->\n\n?','',
                      self._old_read())

    def __repr__(self):
        return ("<%s %s at 0x%s>"
                % (self.__class__.__name__, `self.id()`, hex(id(self))[2:]))

    # methods for reliable dtml access to page & wiki url
    # these have been troublesome; they need to work with & without
    # virtual hosting. Tests needed.
    # Keep old versions around to help with debugging - see
    # http://zwiki.org/VirtualHostingSummary
    # XXX absolute_url uses REQUEST.SERVER_URL - would be nice for these
    # to work without requiring a live request
    
    def page_url(self):
        """return the url path for this wiki page"""
        return self.page_url7()

    def wiki_url(self):
        """return the base url path for this wiki"""
        return self.wiki_url7()

    def page_url1(self):
        """return the url path for the current wiki page"""
        o = self
        url = []
        while hasattr(o,'id'):
            url.insert(0,absattr(o.id))
            o = getattr(o,'aq_parent', None)
        return quote('/' + join(url[1:],'/'))

    def wiki_url1(self):
        """return the base url path for the current wiki"""
        # this code is buried somewhere in cvs
        return '?'

    def page_url2(self):
        """return the url path for the current wiki page"""
        return quote('/' + self.absolute_url(relative=1))

    def wiki_url2(self):
        """return the base url path for the current wiki"""
        return quote('/' + self.folder().absolute_url(relative=1))

    def page_url3(self):
        """return the url path for the current wiki page"""
        return self.REQUEST['BASE1'] + '/' + self.absolute_url(relative=1)

    def wiki_url3(self):
        """return the base url path for the current wiki"""
        return self.REQUEST['BASE1'] + '/' + self.folder().absolute_url(relative=1)

    def page_url4(self):
        """return the url path for the current wiki page"""
        return self.absolute_url(relative=0)

    def wiki_url4(self):
        """return the base url path for the current wiki"""
        return self.folder().absolute_url(relative=0)

    def page_url5(self):
        """return the url path for the current wiki page"""
        return self.absolute_url()

    def wiki_url5(self):
        """return the base url path for the current wiki"""
        return self.folder().absolute_url()

    def page_url6(self):
        """return the url path for the current wiki page"""
        return '/' + self.absolute_url(relative=1)

    def wiki_url6(self):
        """return the base url path for the current wiki"""
        return '/' + self.aq_inner._my_folder_orig().absolute_url(relative=1)

    def _my_folder_orig(self):
        """
        Obtain parent folder, avoiding potential acquisition recursion.

        ken's original version - keep it around in case
        """
        got = parent = self.folder()
        # Handle bizarred degenerate case - darnit, i've forgotten now where
        # i've seen this occur!
        while (got
               and hasattr(got, 'aq_parent')
               and got.aq_parent
               and (got is not got.aq_parent)
               and (aq_base(got) is aq_base(got.aq_parent))):
            parent = got
            got = got.aq_parent
        return parent

    def page_url7(self):
        """return the url path for the current wiki page"""
        return self.wiki_url() + '/' + quote(self.id())

    def wiki_url7(self):
        """return the base url path for the current wiki"""
        try:
            return self.folder().absolute_url()
        except KeyError,AttributeError:
            return '' # makes debugging/testing easier

    def creationTime(self):
        """return our creation time as a DateTime object"""
        if self.creation_time:
            try: return DateTime(self.creation_time)
            except DateTime.SyntaxError: return None
        else:
            return None

    def lastEditTime(self):
        """return our last edit time as a DateTime object"""
        if self.last_edit_time:
            try: return DateTime(self.last_edit_time)
            except DateTime.SyntaxError: return None
        else:
            return None

    def folder(self):
        """
        return this page's containing folder

        We used to use self.aq_parent everywhere, now
        self.aq_inner.aq_parent to ignore acquisition paths.
        Work for pages without a proper acquisition wrapper too.
        """
        return getattr(getattr(self,'aq_inner',self),'aq_parent',None)

    def age(self):
        """
        return a string describing the approximate age of this page
        """
        return self.asAgeString(self.creation_time)

    def lastEditInterval(self):
        """
        return a string describing the approximate interval since last edit
        """
        return self.asAgeString(self.last_edit_time)

    def asAgeString(self,time):
        """
        return a string describing the approximate elapsed period since time

        time may be a DateTime or suitable string. Returns a blank string
        if there was a problem. Based on the dtml version in ZwikiTracker.

        >>> from Products.ZWiki.tests.support import MockZWikiPage
        >>> p = MockZWikiPage().aq_parent.TestPage
        >>> p.asAgeString(p.last_edit_time)
        'some time'
        """
        if not time:
            return 'some time'
        if type(time) is StringType:
            time = DateTime(time)
        elapsed = self.ZopeTime() - time
        hourfactor=0.041666666666666664
        minutefactor=0.00069444444444444447
        secondsfactor=1.1574074074074073e-05
        days=int(math.floor(elapsed))
        weeks=days/7
        months=days/30
        years=days/365
        hours=int(math.floor((elapsed-days)/hourfactor))
        minutes=int(math.floor((elapsed-days-hourfactor*hours)/minutefactor))
        seconds=int(round((
            elapsed-days-hourfactor*hours-minutefactor*minutes)/secondsfactor))
        if years:
            s = "%d year%s" % (years, years > 1 and 's' or '')
        elif months:
            s = "%d month%s" % (months, months > 1 and 's' or '')
        elif weeks:
            s = "%d week%s" % (weeks, weeks > 1 and 's' or '')
        elif days:
            s = "%d day%s" % (days, days > 1 and 's' or '')
        elif hours:
            s = "%d hour%s" % (hours, hours > 1 and 's' or '')
        elif minutes:
            s = "%d minute%s" % (minutes, minutes > 1 and 's' or '')
        else:
            s = "%d second%s" % (seconds, seconds > 1 and 's' or '')
        return s

    def linkTitle(self,prettyprint=0):
        """
        return a suitable value for the title attribute of links to this page

        with prettyprint=1, format it for use in the standard header.
        """
        #return "last edited by %s on %s" % \
        #       (self.last_editor, self.last_edit_time)
        return self.linkTitleFrom(self.last_edit_time,
                                  self.last_editor,
                                  prettyprint=prettyprint)

    # please clean me up
    def linkTitleFrom(self,last_edit_time=None,last_editor=None,prettyprint=0):
        """
        ǽȹԤΥ󥯥ȥʸ
        make a link title string from these last_edit_time and editor strings
        
        with prettyprint=1, format it for use in the standard header.
        """
        interval = self.asAgeString(last_edit_time)
        interval = replace(interval,'seconds','')
        interval = replace(interval,'second','')
        interval = replace(interval,'minutes','ʬ')
        interval = replace(interval,'minute','ʬ')
        interval = replace(interval,'hours','')
        interval = replace(interval,'hour','')
        interval = replace(interval,'days','')
        interval = replace(interval,'day','')
        interval = replace(interval,'weeks','')
        interval = replace(interval,'week','')
        interval = replace(interval,'months','')
        interval = replace(interval,'month','')
        interval = replace(interval,'years','ǯ')
        interval = replace(interval,'year','ǯ')
        if not prettyprint:
            s = " %s " % (interval)
        else:
            try:
                assert self.REQUEST.AUTHENTICATED_USER.has_permission(
                    'View History',self)
                #XXX do timezone conversion ?
                lastlog = self.lastlog()
                if lastlog: lastlog = ' ('+lastlog+')'
#                s = '<b><u>l</u></b>ast edited <a href="%s/diff" title="show last edit%s" accesskey="l">%s</a> ago' % \
                s = '<a href="%s/diff" title="ǽɽ:  (%s)" accesskey="l"><b>%s</b></a>' % \
                    (self.page_url(), lastlog, interval)
            except:
                s = ' %s  ' % (interval)
        if (last_editor and
            not re.match(r'^[0-9\.\s]*$',last_editor)):
            # escape some things that might cause trouble in an attribute
            editor = re.sub(r'"',r'',last_editor)
            if not prettyprint:
                s = s + " by %s" % (editor)
            else:
                s = s + " by <b>%s</b>" % (editor)
        return s
    
    def manage_edit(self, data, title, REQUEST=None):
        """Do standard manage_edit kind of stuff, using our edit."""
        #self.edit(text=data, title=title, REQUEST=REQUEST, check_conflict=0)
        #I think we need to bypass edit to provide correct permissions
        self._setText(data,REQUEST)
        self._setLastEditor(REQUEST)
        self.reindex_object()
        if REQUEST:
            message="Content changed."
            return self.manage_main(self,REQUEST,manage_tabs_message=message)

    ######################################################################
    # backwards compatibility methods

    def upgradeAll(self,pre_render=1):
        """
        Clear cache, upgrade and pre-render all pages

        Normally pages are upgraded/pre-rendered as needed.  An
        administrator may want to call this, particularly after a zwiki
        upgrade, to minimize later delays and to ensure all pages have
        been rendered by the latest code.

        Requires 'Manage properties' permission on the folder.
        Commit every so often an attempt to avoid memory/conflict errors.
        Has problems doing a complete run in large wikis, or when other
        page accesses are going on ?
        """
        if not self._checkPermission('Manage properties',self.folder()):
            raise 'Unauthorized', (
             _('You are not authorized to upgrade all pages.') + \
             _('(folder -> Manage properties)'))

        if pre_render:
            DLOG('upgrading and prerendering all pages:')
        else:
            DLOG('upgrading all pages:')
        n = 0
        for p in self.folder().objectValues(spec=self.meta_type):
            n = n + 1
            p.upgrade()
            if pre_render:
                p._preRender(clear_cache=1)
            DLOG('page #%d %s'%(n,p.id()))
            if n % 10 == 0:
                DLOG('committing')
                get_transaction().commit()
            # last pages will get committed as this request ends
        DLOG('%d pages processed' % n)

    def upgrade(self):
        """
        Upgrade an old page instance (and possibly the parent folder).

        Called as needed, ie at view time (set AUTO_UPGRADE=0 in
        Default.py to prevent this).  You could also call this on every
        page in your wiki to do a batch upgrade. Affects
        bobobase_modification_time. If you later downgrade zwiki, the
        upgraded pages may not work so well.
        """
        # Note that the objects don't get very far unpickling, some
        # by-hand adjustment via command-line interaction is necessary
        # to get them over the transition, sigh. --ken
        # not sure what this means --SM

        # What happens in the zodb when class definitions change ? I think
        # all instances in the zodb conform to the new class shape
        # immediately on refresh/restart, but what happens to
        # (a) old _properties lists ? not sure, assume they remain in
        # unaffected and we need to add the new properties
        # and (b) old properties & attributes no longer in the class
        # definition ?  I think these lurk around, and we need to delete
        # them.

        # fix up attributes first, then properties
        changed = 0

        # don't acquire during surgery..
        realself = self
        self = self.aq_base

        # upgrade old page types
        # choose pre-formatting & pre-linking variants where possible
        changedpagetypes = {
            # early zwiki
            'Structured Text'    :'stxprelinkhtml',
            'structuredtext_dtml':'stxprelinkdtmlhtml',
            'HTML'               :'prelinkhtml',
            'html_dtml'          :'prelinkdtmlhtml',
            'Classic Wiki'       :'wwmlprelink',
            'Plain Text'         :'plaintext',
            # pre-0.9.10
            'stxprelinkdtml'     :'stxprelinkdtmlhtml',
            'structuredtextdtml' :'stxprelinkdtmlhtml',
            'dtmlstructuredtext' :'stxprelinkdtmlhtml',
            'structuredtext'     :'stxprelinkhtml',
            'structuredtextonly' :'stxprelink',
            'classicwiki'        :'wwmlprelink',
            'htmldtml'           :'prelinkdtmlhtml',
            # I want to reuse the 'html' type, so leave it be.  Old 'html'
            # pages will lose their wikilinks; add to release notes and
            # let people fix manually.
            'plainhtmldtml'      :'dtmlhtml',
            'plainhtml'          :'html',
            }
        if self.page_type in changedpagetypes.keys():
            self.page_type = changedpagetypes[self.page_type]
            # clear render cache; don't bother prerendering just now
            self.clearCache()
            changed = 1

        # Early zwikipages have a username string property - delete it.
        # Once it's gone, the username method below is revealed and allows
        # old dtml to keep working. 
        # Could attempt to transfer it to last_editor, then you'd need
        # to transfer timestamp as well or it looks confusing. Don't bother.
        if type(self.username) is StringType: 
            delattr(self,'username')
            changed = 1 

        # Pre-0.9.10, creation_time has been a string in custom format and
        # last_edit_time has been a DateTime. Now both are kept as
        # ISO-format strings. Might not be strictly necessary to upgrade
        # in all cases.. will cause a lot of bobobase_mod_time
        # updates.. do it anyway.
        if not self.last_edit_time:
            self.last_edit_time = self.bobobase_modification_time().ISO()
            changed = 1
        elif type(self.last_edit_time) is not StringType:
            self.last_edit_time = self.last_edit_time.ISO()
            changed = 1
        elif len(self.last_edit_time) != 19:
            try: 
                self.last_edit_time = DateTime(self.last_edit_time).ISO()
                changed = 1
            except DateTime.SyntaxError:
                # can't convert to ISO, just leave it be
                pass

        # If no creation_time, just leave it blank for now.
        # we shouldn't find DateTimes here, but check anyway
        if not self.creation_time:
            pass
        elif type(self.creation_time) is not StringType:
            self.creation_time = self.creation_time.ISO()
            changed = 1
        elif len(self.creation_time) != 19:
            try: 
                self.creation_time = DateTime(self.creation_time).ISO()
                changed = 1
            except DateTime.SyntaxError:
                # can't convert to ISO, just leave it be
                pass

        # _wikilinks is now _links
        #if hasattr(self.aq_base,'_wikilinks'): #XXX why doesn't this work
        if hasattr(self,'_wikilinks'):
            self._links = self._wikilinks.keys()[:]
            delattr(self,'_wikilinks')
            changed = 1 

        # update _properties
        # keep in sync with _properties above. Better if we used that as
        # the template (efficiently)
        oldprops = { # not implemented
            'page_type'     :{'id':'page_type','type':'string'},
            }
        newprops = {
            #'page_type'     :{'id':'page_type','type':'selection','mode': 'w',
            #                  'select_variable': 'ZWIKI_PAGE_TYPES'},
            'creator'       :{'id':'creator','type':'string','mode':'r'},
            'creator_ip'    :{'id':'creator_ip','type':'string','mode':'r'},
            'creation_time' :{'id':'creation_time','type':'string','mode':'r'},
            'last_editor'   :{'id':'last_editor','type':'string','mode':'r'},
            'last_editor_ip':{'id':'last_editor_ip','type':'string','mode':'r'},
            'last_edit_time':{'id':'creation_time','type':'string','mode':'r'},
            'last_log'      :{'id':'last_log', 'type': 'string', 'mode': 'r'},
            'NOT_CATALOGED' :{'id':'NOT_CATALOGED', 'type': 'boolean', 'mode': 'w'},
            }
        props = map(lambda x:x['id'], self._properties)
        for p in oldprops.keys():
            if p in props: # and oldprops[p]['type'] != blah blah blah :
                pass
                #ack!
                #self._properties = filter(lambda x:x['id'] != p,
                #                          self._properties)
                #changed = 1
        for p in newprops.keys():
            if not p in props:
                self._properties = self._properties + (newprops[p],)
                changed = 1

        if changed:
            # bobobase_modification_time changed - put in a dummy user so
            # it's clear this was not an edit
            # no - you should be looking at last_edit_times, in which case
            # you don't want to see last_editor change for this.
            #self.last_editor_ip = ''
            #self.last_editor = 'UpGrade'
            # do a commit now so the current render will have the
            # correct bobobase_modification_time for display (many
            # headers/footers still show it)
            get_transaction().commit()
            # and log it
            DLOG('upgraded '+self.id())

        # finally, MailSupport does a bit more (merge here ?)
        realself.upgradeSubscribers()

    # some CMF compatibility methods for standard ZWikiPage
    def inCMF(self):
        """return true if this page is in a CMF portal"""
        return hasattr(self.aq_inner.aq_parent,'portal_membership')

    SearchableText = text
    view = __call__
            
    # old API methods to help keep legacy DTML working
    # __roles__=None required so they can be accessed without a docstring ?
    wiki_page_url = page_url
    wiki_page_url__roles__ = None
    wiki_base_url = wiki_url
    wiki_base_url__roles__ = None
    editTimestamp = timeStamp
    editTimestamp__roles__ = None
    checkEditTimeStamp = checkEditConflict
    checkEditTimeStamp__roles__ = None
    doLegacyFixups = upgrade
    #and what if I resist this mindless None'ing of roles ?
    #doLegacyFixups__roles__ = None
    doSubscriberListFixups = MailSupport.upgradeSubscribers
    #doSubscriberListFixups__roles__ = None
    doSubscriberFixups = MailSupport.upgradeSubscribers
    #doSubscriberFixups__roles__ = None
    # old render methods - these shouldn't be needed after upgrade
    # converts page types, but it may be disabled
    render_stxprelinkdtml     = render_stxprelinkdtmlhtml
    render_structuredtextdtml = render_stxdtmllinkhtml
    render_dtmlstructuredtext = render_dtmlstxlinkhtml
    render_structuredtext     = render_stxlinkhtml
    render_structuredtextonly = render_stxlink
    render_classicwiki        = render_wwmllink
    render_htmldtml           = render_dtmllinkhtml
    render_plainhtmldtml      = render_dtmlhtml
    render_plainhtml          = render_html
    
    def username(self):
        # backwards compatibility for dtml which tries to display the old
        # username property. NB upgrade may not have been called yet.
        # XXX I think we can be less slavishly compatible at this
        # point. Just return last_editor ?
        for p in self._properties:
            if p['id'] == 'username':
                return username
        if hasattr(self,'last_editor') and self.last_editor:
            return self.last_editor
        if hasattr(self,'last_editor_ip') and self.last_editor_ip:
            return self.last_editor_ip
        return ''
        
    # actually the above is useful to keep around as
    last_editor_or_ip = username
    last_editor_or_ip__roles__ = None

    # document_src is the standard zope accessor for raw content
    # inherited from DTMLDocument
    # text/src is a permission-free accessor for the content
    # XXX still need this ? document_src & src equivalent ? clean up
    src = text
    src__roles__ = None

    def wikilink(self, text):
        """
        api utility method for wiki-linking an arbitrary text snippet
        """
        #return self._oldRenderLinksIn(text,doLineEscapes=1)
        return self.renderLinksIn(text,doLineEscapes=1)


######################################################################
# FUNCTION CATEGORY: ZMI
######################################################################

# ZMI page creation form
manage_addZWikiPageForm = HTMLFile('dtml/zwikiPageAdd', globals())

def manage_addZWikiPage(self, id, title='', file='', REQUEST=None,
                        submit=None):
    """
    Add a ZWiki Page object with the contents of file.

    Usually zwiki pages are created by clicking on ? (via create); this
    allows them to be added in the standard ZMI way. These should give
    mostly similar results; refactor the two methods together if possible.

    >>> from Products.ZWiki.ZWikiPage import manage_addZWikiPage
    >>> from OFS.Folder import Folder
    >>> f = Folder()
    >>> manage_addZWikiPage(f,'ZmiTestPage')
    ''
    >>> assert hasattr(f,'ZmiTestPage')

    """
    # create page and initialize in proper order, as in create.
    p = ZWikiPage(source_string='', __name__=id)
    newid = self._setObject(id,p)
    p = getattr(self,newid)
    p.title=title
    p._setCreator(REQUEST)
    p._setLastEditor(REQUEST)
    p._setOwnership(REQUEST)
    if hasattr(self,'standard_page_type'):
        p.page_type = self.standard_page_type
    else:
        p.page_type = DEFAULT_PAGE_TYPE
    text = file
    if type(text) is not StringType: text=text.read()
    p._setText(text or '',REQUEST)
    p.index_object()
    if p.usingRegulations():
        p.setRegulations(REQUEST,new=1)
    if REQUEST is not None:
        try: u=self.DestinationURL()
        except: u=REQUEST['URL1']
        if submit==" Add and Edit ": u="%s/%s" % (u,quote(id))
        REQUEST.RESPONSE.redirect(u+'/manage_main')
    return ''


#python notes
#
#   "The first line should always be a short, concise summary of the
#object's purpose.  For brevity, it should not explicitly state the
#object's name or type, since these are available by other means (except
#if the name happens to be a verb describing a function's operation).
#This line should begin with a capital letter and end with a period.
#
#   If there are more lines in the documentation string, the second line
#should be blank, visually separating the summary from the rest of the
#description.  The following lines should be one or more paragraphs
#describing the object's calling conventions, its side effects, etc."
#
#   "Data attributes override method attributes with the same name; to
#avoid accidental name conflicts, which may cause hard-to-find bugs in
#large programs, it is wise to use some kind of convention that
#...
#verbs for methods and nouns for data attributes."
